#!/usr/bin/env python

# This example demonstrates the use of vtkXYPlotActor to display three
# probe lines using three different techniques.  In this example, we
# are loading data using the vtkPLOT3DReader.  We are using the
# vtkProbeFilter to extract the underlying point data along three
# probe lines.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Create a PLOT3D reader and load the data.
pl3d = vtk.vtkPLOT3DReader()
pl3d.SetXYZFileName(VTK_DATA_ROOT + "/Data/combxyz.bin")
pl3d.SetQFileName(VTK_DATA_ROOT + "/Data/combq.bin")
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.Update()

# Create three the line source to use for the probe lines.
line = vtk.vtkLineSource()
line.SetResolution(30)

# Move the line into place and create the probe filter.  For
# vtkProbeFilter, the probe line is the input, and the underlying data
# set is the source.
transL1 = vtk.vtkTransform()
transL1.Translate(3.7, 0.0, 28.37)
transL1.Scale(5, 5, 5)
transL1.RotateY(90)
tf = vtk.vtkTransformPolyDataFilter()
tf.SetInputConnection(line.GetOutputPort())
tf.SetTransform(transL1)
probe = vtk.vtkProbeFilter()
probe.SetInputConnection(tf.GetOutputPort())
probe.SetSource(pl3d.GetOutput())

# Move the line again and create another probe filter.
transL2 = vtk.vtkTransform()
transL2.Translate(9.2, 0.0, 31.20)
transL2.Scale(5, 5, 5)
transL2.RotateY(90)
tf2 = vtk.vtkTransformPolyDataFilter()
tf2.SetInputConnection(line.GetOutputPort())
tf2.SetTransform(transL2)
probe2 = vtk.vtkProbeFilter()
probe2.SetInputConnection(tf2.GetOutputPort())
probe2.SetSource(pl3d.GetOutput())

# Move the line again and create a third probe filter.
transL3 = vtk.vtkTransform()
transL3.Translate(13.27, 0.0, 33.40)
transL3.Scale(4.5, 4.5, 4.5)
transL3.RotateY(90)
tf3 = vtk.vtkTransformPolyDataFilter()
tf3.SetInputConnection(line.GetOutputPort())
tf3.SetTransform(transL3)
probe3 = vtk.vtkProbeFilter()
probe3.SetInputConnection(tf3.GetOutputPort())
probe3.SetSource(pl3d.GetOutput())

# Create a vtkAppendPolyData to merge the output of the three probe
# filters into one data set.
appendF = vtk.vtkAppendPolyData()
appendF.AddInput(probe.GetPolyDataOutput())
appendF.AddInput(probe2.GetPolyDataOutput())
appendF.AddInput(probe3.GetPolyDataOutput())

# Create a tube filter to represent the lines as tubes.  Set up the
# associated mapper and actor.
tuber = vtk.vtkTubeFilter()
tuber.SetInputConnection(appendF.GetOutputPort())
tuber.SetRadius(0.1)
lineMapper = vtk.vtkPolyDataMapper()
lineMapper.SetInputConnection(tuber.GetOutputPort())
lineActor = vtk.vtkActor()
lineActor.SetMapper(lineMapper)

# Create an xy-plot using the output of the 3 probe filters as input.
# The x-values we are plotting are arc length.
xyplot = vtk.vtkXYPlotActor()
xyplot.AddInput(probe.GetOutput())
xyplot.AddInput(probe2.GetOutput())
xyplot.AddInput(probe3.GetOutput())
xyplot.GetPositionCoordinate().SetValue(0.0, 0.67, 0)
xyplot.GetPosition2Coordinate().SetValue(1.0, 0.33, 0) #relative to Position
xyplot.SetXValuesToArcLength()
xyplot.SetNumberOfXLabels(6)
xyplot.SetTitle("Pressure vs. Arc Length (Zoomed View)")
xyplot.SetXTitle("")
xyplot.SetYTitle("P")
xyplot.SetXRange(.1, .35)
xyplot.SetYRange(.2, .4)
xyplot.GetProperty().SetColor(0, 0, 0)
xyplot.GetProperty().SetLineWidth(2)
# Set text prop color (same color for backward compat with test)
# Assign same object to all text props
tprop = xyplot.GetTitleTextProperty()
tprop.SetColor(xyplot.GetProperty().GetColor())
xyplot.SetAxisTitleTextProperty(tprop)
xyplot.SetAxisLabelTextProperty(tprop)

# Create an xy-plot using the output of the 3 probe filters as input.
# The x-values we are plotting are normalized arc length.
xyplot2 = vtk.vtkXYPlotActor()
xyplot2.AddInput(probe.GetOutput())
xyplot2.AddInput(probe2.GetOutput())
xyplot2.AddInput(probe3.GetOutput())
xyplot2.GetPositionCoordinate().SetValue(0.00, 0.33, 0)
xyplot2.GetPosition2Coordinate().SetValue(1.0, 0.33, 0) #relative to Position
xyplot2.SetXValuesToNormalizedArcLength()
xyplot2.SetNumberOfXLabels(6)
xyplot2.SetTitle("Pressure vs. Normalized Arc Length")
xyplot2.SetXTitle("")
xyplot2.SetYTitle("P")
xyplot2.PlotPointsOn()
xyplot2.PlotLinesOff()
xyplot2.GetProperty().SetColor(1, 0, 0)
xyplot2.GetProperty().SetPointSize(2)
# Set text prop color (same color for backward compat with test)
# Assign same object to all text props
tprop = xyplot2.GetTitleTextProperty()
tprop.SetColor(xyplot2.GetProperty().GetColor())
xyplot2.SetAxisTitleTextProperty(tprop)
xyplot2.SetAxisLabelTextProperty(tprop)

# Create an xy-plot using the output of the 3 probe filters as input.
# The x-values we are plotting are the underlying point data values.
xyplot3 = vtk.vtkXYPlotActor()
xyplot3.AddInput(probe.GetOutput())
xyplot3.AddInput(probe2.GetOutput())
xyplot3.AddInput(probe3.GetOutput())
xyplot3.GetPositionCoordinate().SetValue(0.0, 0.0, 0)
xyplot3.GetPosition2Coordinate().SetValue(1.0, 0.33, 0) #relative to Position
xyplot3.SetXValuesToIndex()
xyplot3.SetNumberOfXLabels(6)
xyplot3.SetTitle("Pressure vs. Point Id")
xyplot3.SetXTitle("Probe Length")
xyplot3.SetYTitle("P")
xyplot3.PlotPointsOn()
xyplot3.GetProperty().SetColor(0, 0, 1)
xyplot3.GetProperty().SetPointSize(3)
# Set text prop color (same color for backward compat with test)
# Assign same object to all text props
tprop = xyplot3.GetTitleTextProperty()
tprop.SetColor(xyplot3.GetProperty().GetColor())
xyplot3.SetAxisTitleTextProperty(tprop)
xyplot3.SetAxisLabelTextProperty(tprop)

# Draw an outline of the PLOT3D data set.
outline = vtk.vtkStructuredGridOutlineFilter()
outline.SetInputConnection(pl3d.GetOutputPort())
outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outline.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)
outlineActor.GetProperty().SetColor(0, 0, 0)

# Create the Renderers, RenderWindow, and RenderWindowInteractor.
ren = vtk.vtkRenderer()
ren2 = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
renWin.AddRenderer(ren2)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Set the background, viewport (necessary because we want to have the
# renderers draw to different parts of the render window) of the first
# renderer.  Add the outline and line actors to the renderer.
ren.SetBackground(0.6784, 0.8471, 0.9020)
ren.SetViewport(0, 0, .5, 1)
ren.AddActor(outlineActor)
ren.AddActor(lineActor)

# Set the background and viewport of the second renderer.  Add the
# xy-plot actors to the renderer.  Set the size of the render window.
ren2.SetBackground(1, 1, 1)
ren2.SetViewport(0.5, 0.0, 1.0, 1.0)
ren2.AddActor2D(xyplot)
ren2.AddActor2D(xyplot2)
ren2.AddActor2D(xyplot3)
renWin.SetSize(500, 250)

# Set up the camera parameters.
cam1 = ren.GetActiveCamera()
cam1.SetClippingRange(3.95297, 100)
cam1.SetFocalPoint(8.88908, 0.595038, 29.3342)
cam1.SetPosition(-12.3332, 31.7479, 41.2387)
cam1.SetViewUp(0.060772, -0.319905, 0.945498)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
import pydoc

from vistrails.core import debug
from vistrails.core.utils import VistrailsInternalError
from vistrails.core.vistrail.port_spec import PortSpec
import vistrails.core.modules.module_registry
from vistrails.core.modules.utils import create_descriptor_string
from vistrails.db.domain import DBModuleDescriptor

# this is used by add_port to signal a repeated port. Should never
# happen, but it does. Probably means a bug on the dynamic modules
# such as MplPlot and Tuple.  Of course, that also means that we
# should be robust in the presence of these errors, since user-defined
# modules could exhibit this bug

class OverloadedPort(Exception):
    pass

###############################################################################
# ModuleDescriptor

class ModuleDescriptor(DBModuleDescriptor):
    """ModuleDescriptor is a class that holds information about
    modules in the registry. There exists exactly one ModuleDescriptor
    for every registered VisTrails module in the system.

    self.module: reference to the python class that defines the module
    self.name: name of the module
    self.identifier: identifier of the package that module belongs to
    self.input_ports: dictionary of names of input ports to the types
      consumed by the ports
    self.output_ports: dictionary of names of output ports to the types
      produces by the ports
    self.input_ports_optional: dictionary of input port names that records
      whether ports should show up by default on GUI
    self.output_ports_optional: dictionary of output port names that records
      whether ports should show up by default on GUI
    self.port_order: stores a map from names to numbers to order the ports
      in the GUI

    self._is_abstract: whether module is abstract
    self._configuration_widget: reference to the Qt class that provides a
      custom configuration widget for the class.  Note that this can
      be a tuple (path, name) that will be loaded only when needed via
      __import__ (! this is preferred !) or as a QWidget (deprecated)

    self._left_fringe and self._right_fringe: lists of 2D points that
      define a drawing style for modules in the GUI
    self._module_color: color of the module in the GUI

    self._widget_item: stores a reference to the ModuleTreeWidgetItem so
      that when ports are added to modules things get correctly updated.

    self._input_port_cache, self._output_port_cache,
      self._port_caches: Dictionaries for fast port spec lookup,
      created because port spec lookups are sometimes part of hot code
      paths and need to go as fast as possible.
    """

    ##########################################################################

    def __init__(self, *args, **kwargs):
        self.children = []
        if 'module' in kwargs:
            self.module = kwargs['module']
            if 'name' not in kwargs:
                kwargs['name'] = self.module.__name__
            del kwargs['module']
        else:
            self.module = None
        if 'base_descriptor' in kwargs:
            if kwargs['base_descriptor'] is not None:
                self._base_descriptor = kwargs['base_descriptor']
                if 'base_descriptor_id' not in kwargs:
                    kwargs['base_descriptor_id'] = self._base_descriptor.id
                self._port_count = self._base_descriptor._port_count
                self._base_descriptor.children.append(self)
            else:
                self._base_descriptor = None
                self._port_count = 0
            del kwargs['base_descriptor']
        else:
            self._base_descriptor = None
            self._port_count = 0
        # rename identifier to package for db
        if 'identifier' in kwargs:
            if 'package' not in kwargs:
                kwargs['package'] = kwargs['identifier']
            del kwargs['identifier']
        if 'base_descriptor_id' not in kwargs:
            kwargs['base_descriptor_id'] = -1
        DBModuleDescriptor.__init__(self, *args, **kwargs)
        self.set_defaults()

    def set_defaults(self, other=None):
        if other is None:
            self._abstraction_refs = 1
            self._is_abstract = False
            self._configuration_widget = None
            self._left_fringe = None
            self._right_fringe = None
            self._module_color = None
            self._hasher_callable = None
            self._widget_item = None
            self._is_hidden = False
            self._namespace_hidden = False
            self._widget_classes = {}
            self.children = []
            # The ghost attributes represent the original values
            # for the descriptor of an upgraded package subworkflow
            # so it can be displayed with it's original package
            # rather than 'local.abstractions'
            self.ghost_identifier = ''
            self.ghost_package_version = ''
            self.ghost_namespace = None
        else:
            # FIXME this will break things, I think
            self.children = copy.copy(other.children)
            
            self._base_descriptor = other._base_descriptor
            self.module = other.module
            self._port_count = other._port_count
            self._abstraction_refs = self._abstraction_refs
            self._is_abstract = other._is_abstract
            self._configuration_widget = other._configuration_widget
            self._left_fringe = other._left_fringe
            self._right_fringe = other._right_fringe
            self._module_color = other._module_color
            self._hasher_callable = other._hasher_callable
            self._widget_item = other._widget_item
            self._is_hidden = other._is_hidden
            self._widget_classes = dict((k,copy.copy(v)) for k, v in \
                                         other._widget_classes.iteritems())
            self._namespace_hidden = other._namespace_hidden
            self.ghost_identifier = other.ghost_identifier
            self.ghost_package_version = other.ghost_package_version
            self.ghost_namespace = other.ghost_namespace
        self.port_specs = self.db_portSpecs_name_index
        if self.version is None:
            self.version = ''
        if self.namespace is None:
            self.namespace = ''

    def __copy__(self):
        return ModuleDescriptor.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBModuleDescriptor.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ModuleDescriptor
        cp.set_defaults(self)
        return cp
        
    @staticmethod
    def convert(_desc):
        if _desc.__class__ == ModuleDescriptor:
            return
        _desc.__class__ = ModuleDescriptor
        
        for port_spec in _desc.db_portSpecs:
            PortSpec.convert(port_spec)

        # do more init stuff
        _desc.children = []
        _desc.module = None
        _desc._base_descriptor = None
        _desc._port_count = 0
        _desc.set_defaults()

    ##########################################################################
    # Properties

    id = DBModuleDescriptor.db_id
    name = DBModuleDescriptor.db_name
    identifier = DBModuleDescriptor.db_package
    package = DBModuleDescriptor.db_package
    namespace = DBModuleDescriptor.db_namespace
    package_version = DBModuleDescriptor.db_package_version
    version = DBModuleDescriptor.db_version
    base_descriptor_id = DBModuleDescriptor.db_base_descriptor_id
    port_specs_list = DBModuleDescriptor.db_portSpecs
    
    def _get_base_descriptor(self):
        if self._base_descriptor is None and self.base_descriptor_id >= 0:
            from vistrails.core.modules.module_registry import get_module_registry
            reg = get_module_registry()
            self._base_descriptor = \
                reg.descriptors_by_id[self.base_descriptor_id]
        return self._base_descriptor
    def _set_base_descriptor(self, base_descriptor):
        self._base_descriptor = base_descriptor
        self.base_descriptor_id = base_descriptor.id
    base_descriptor = property(_get_base_descriptor, _set_base_descriptor)

    def _get_sigstring(self):
        return create_descriptor_string(self.db_package, self.db_name,
                                        self.db_namespace)
    sigstring = property(_get_sigstring)

    def _get_spec_tuple(self):
        return (self.package, self.name, self.namespace)
    spec_tuple = property(_get_spec_tuple)

    def set_module_abstract(self, v):
        self._is_abstract = v

    def module_abstract(self):
#         if not self.has_ports():
#             return True
        return self._is_abstract

    def set_configuration_widget(self, configuration_widget_type):
        """set_configuration_widget(configuration_widget_type: 
                                        (path, name) tuple or QWidget) -> None
        
        """
        self._configuration_widget = configuration_widget_type

    def configuration_widget(self):
        return self._configuration_widget

    def set_constant_config_widget(self, widget_class, widget_use, 
                                   widget_type):
        if widget_use not in self._widget_classes:
            self._widget_classes[widget_use] = {}
        self._widget_classes[widget_use][widget_type] = widget_class

    def has_constant_config_widget(self, widget_use, widget_type):
        return widget_use in self._widget_classes and \
            widget_type in self._widget_classes[widget_use]

    def get_constant_config_widget(self, widget_use, widget_type):
        if self.has_constant_config_widget(widget_use, widget_type):
            return self._widget_classes[widget_use][widget_type]
        return None

    def get_all_constant_config_widgets(self, widget_use):
        if widget_use in self._widget_classes:
            return self._widget_classes[widget_use]
        return {}

    def set_module_color(self, color):
        if color:
            assert isinstance(color, tuple)
            assert len(color) == 3
            for i in 0,1,2:
                assert isinstance(color[i], float)
        self._module_color = color

    def module_color(self):
        return self._module_color

    def set_module_fringe(self, left_fringe, right_fringe):
        if left_fringe is None:
            assert right_fringe is None
            self._left_fringe = None
            self._right_fringe = None
        else:
            vistrails.core.modules.module_registry._check_fringe(left_fringe)
            vistrails.core.modules.module_registry._check_fringe(right_fringe)
            self._left_fringe = left_fringe
            self._right_fringe = right_fringe

    def module_fringe(self):
        if self._left_fringe is None and self._right_fringe is None:
            return None
        return (self._left_fringe, self._right_fringe)

    def module_documentation(self, module=None):
        doc = pydoc.getdoc(self.module)
        if hasattr(self.module, 'get_documentation'):
            try:
                doc = self.module.get_documentation(doc, module)
            except Exception, e:
                debug.critical("Exception calling get_documentation on %r" %
                               self.module,
                               e)
                doc = doc or "(Error getting documentation)"
        doc = doc or "(No documentation available)"
        return doc

    def module_package(self):
        return self.identifier

    def set_hasher_callable(self, callable_):
        self._hasher_callable = callable_
    def hasher_callable(self):
        return self._hasher_callable

    def _get_is_hidden(self):
        return self._is_hidden
    def _set_is_hidden(self, hidden):
        self._is_hidden = hidden
    is_hidden = property(_get_is_hidden, _set_is_hidden)

    def _get_namespace_hidden(self):
        return self._namespace_hidden
    def _set_namespace_hidden(self, hidden):
        self._namespace_hidden = hidden
    namespace_hidden = property(_get_namespace_hidden, _set_namespace_hidden)

    ##########################################################################
    # Operators

    def __hash__(self):
        return (type(self), self.package, self.name, self.namespace, 
                self.version).__hash__()

    def __eq__(self, other):
        return (self.package == other.package and
                self.name == other.name and
                self.namespace == other.namespace and
                self.version == other.version)

    def __str__(self):
        return ("ModuleDescriptor(id=%s, package=%s, name=%s, namespace=%s, "
                "version=%s, base_descriptor_id=%s)" % \
                    (self.id, self.package, self.name, self.namespace,
                     self.version, self.base_descriptor_id))

    ##########################################################################
    # Abstract module detection support

    def set_widget(self, widget_item):
        self._widget_item = widget_item

    def has_ports(self):
        """Returns True is module has any ports (this includes
        superclasses).  This method exists to make automatic abstract
        module detection efficient."""
        return self._port_count > 0

    def port_count(self):
        """Return the total number of available for the module."""
        return self._port_count

    # Signal handling
    def new_input_port(self):
        """Updates needed variables when new input port is added
        to either this module or the superclass."""
        self._port_count += 1
        if self._widget_item:
            self._widget_item.added_input_port()
        for child in self.children:
            child.new_input_port()
        
    def new_output_port(self):
        """Updates needed variables when new output port is added
        to either this module or the superclass."""
        self._port_count += 1
        if self._widget_item:
            self._widget_item.added_output_port()
        for child in self.children:
            child.new_output_port()

    ##########################################################################

    # port_type is 'input' or 'output'
    def has_port_spec(self, name, port_type):
        return self.db_has_portSpec_with_name((name, port_type))

    def get_port_spec(self, name, port_type):
        if not self.db_has_portSpec_with_name((name, port_type)):
            raise ValueError("ModuleDescriptor.get_port_spec called when spec "
                             " (%s, %s) doesn't exist" % (name, port_type))
        return self.db_get_portSpec_by_name((name, port_type))

    def set_port_spec(self, name, port_type, port_spec):
        self.db_add_portSpec(port_spec)

    def add_port_spec(self, port_spec):
        self.db_add_portSpec(port_spec)

    def delete_port_spec(self, port_spec):
        self.db_delete_portSpec(port_spec)
        
    def delete_input_port(self, name):
        key = (name, 'input')
        if key in self.port_specs:
            self.delete_port_spec(self.port_specs[key])
        else:
            raise VistrailsInternalError("delete_input_port called on "
                                         "nonexistent port '%s'" % name)

    def delete_output_port(self, name):
        key = (name, 'output')
        if key in self.port_specs:
            self.delete_port_spec(self.port_specs[key])
        else:
            raise VistrailsInternalError("delete_output_port called on "
                                         "nonexistent port '%s'" % name)






###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from vistrails.gui.application import get_vistrails_application

import unittest
import copy
import random
import vistrails.gui.utils

##############################################################################
# Exceptions

class NoVistrail(Exception):
    pass

class NoGUI(Exception):
    pass

##############################################################################

def switch_to_pipeline_view():
    """switch_to_pipeline_view():

    Changes current viewing mode to pipeline view in the builder window.

    """
    get_vistrails_application().builderWindow.qactions['pipeline'].trigger()

def switch_to_history_view():
    """switch_to_history_view():

    Changes current viewing mode to history view in the builder window.

    """
    get_vistrails_application().builderWindow.qactions['history'].trigger()
    
def switch_to_query_view():
    """switch_to_query_view():

    Changes current viewing mode to query view in the builder window.

    """
    get_vistrails_application().builderWindow.qactions['search'].trigger()

def switch_to_mashup_view():
    """switch_to_mashup_view():

    Changes current viewing mode to mashup view in the builder window.

    """
    get_vistrails_application().builderWindow.qactions['mashup'].trigger()
 
################################################################################
# Access to current state

def get_builder_window():
    """get_builder_window():

    returns the main VisTrails GUI window

    raises NoGUI.

    """
    try:
        return get_vistrails_application().builderWindow
    except AttributeError:
        raise NoGUI
    
def get_current_controller():
    """get_current_controller():

    returns the VistrailController of the currently selected vistrail.

    raises NoVistrail.

    """
    try:
        return get_vistrails_application().builderWindow.get_current_controller()
    except AttributeError:
        raise NoVistrail

def get_current_vistrail():
    """get_current_vistrail():

    Returns the currently selected vistrail.

    """
    return get_current_controller().vistrail

def get_current_vistrail_view():
    """get_current_vistrail():

    Returns the currently selected vistrail view.

    """
    view = get_vistrails_application().builderWindow.get_current_view()
    if view is None:
        raise NoVistrail
    return view    

def close_current_vistrail(quiet=False):
    get_vistrails_application().builderWindow.close_vistrail(get_current_vistrail_view())

def get_module_registry():
    from vistrails.core.modules.module_registry import get_module_registry
    return get_module_registry()

##############################################################################
# Do things

def add_module(x, y, identifier, name, namespace, controller=None):
    if controller is None:
        controller = get_current_controller()
    if controller.current_version==-1:
        controller.change_selected_version(0)
    result = controller.add_module(x, y, identifier, name, namespace)
    controller.updatePipelineScene()
    result = controller.current_pipeline.modules[result.id]
    return result
    
def add_module_from_descriptor(descriptor, x=0.0, y=0.0, 
                               internal_version=-1, controller=None):
    if controller is None:
        controller = get_current_controller()
    if controller.current_version==-1:
        controller.change_selected_version(0)
    result = controller.add_module_from_descriptor(descriptor, x, y, 
                                                   internal_version)
    controller.updatePipelineScene()
    result = controller.current_pipeline.modules[result.id]
    return result
    
def add_connection(output_id, output_port_spec, input_id, input_port_spec, 
                   controller=None):
    if controller is None:
        controller = get_current_controller()
    result = controller.add_connection(output_id, output_port_spec,
                                       input_id, input_port_spec)
    controller.updatePipelineScene()
    result = controller.current_pipeline.connections[result.id]
    return result

def create_group(module_ids, connection_ids, controller=None):
    if controller is None:
        controller = get_current_controller()
    group = controller.create_group(module_ids, connection_ids)
    controller.updatePipelineScene()
    return group

def get_modules_by_name(name, package=None, namespace=None, controller=None):
    if controller is None:
        controller = get_current_controller()
    res = []
    for module in controller.current_pipeline.modules.itervalues():
        if (module.name == name and
            (package is None or module.package == package) and
            (namespace is None or module.namespace == namespace)):
            res.append(module)
    return res

def get_selected_modules(controller=None):
    if controller is None:
        controller = get_current_controller()
    modules = []
    selected = controller.get_selected_item_ids()
    if selected is None:
        return []
    (sel_module_ids, sel_connection_ids) = selected
    for m_id in sel_module_ids:
        modules.append(controller.current_pipeline.modules[m_id])
    return modules
    
def change_parameter(module_id, function_name, param_list, function_id=-1L,
                     alias_list=[], controller=None):
    """change_parameter(module_id: long, 
                        function_name: str, 
                        param_list: list(str),
                        function_id: long,
                        alias_list: list(str),
                        controller: VistrailController,
                        ) -> None
    Note: param_list is a list of strings no matter what the parameter type!
    Note: alias_list will be REMOVED!!
    """
    if controller is None:
        controller = get_current_controller()
    module = controller.current_pipeline.modules[module_id]
    controller.update_function(module, function_name, param_list, function_id, 
                               alias_list)
    controller.updatePipelineScene()

def change_parameter_by_id(module_id, function_id, old_param_id, new_value, 
                           controller=None):
    """change_parameter_by_id(module_id: long,
                             function_id: long,
                             old_param_id: long,
                             new_value: str,
                             controller: VistrailController) -> long
    Returns the id of the new parameter.
    Note: function_id is the real_id! Use f.real_id to access real_id
    Note: old_param_id is the real_id! Use p.real_id to access real_id
    """
    if controller is None:
        controller = get_current_controller()
    module = controller.current_pipeline.modules[module_id]
    function = module.function_idx[function_id]
    pos = function.parameter_idx[old_param_id].pos
    controller.update_parameter(function, old_param_id, new_value)
    controller.updatePipelineScene()
    return function.params[pos].real_id

def change_parameter_by_pos(module_id, function_pos, old_param_pos, new_value,
                            controller=None):
    """change_parameter_by_id(module_id: long,
                             function_pos: int,
                             old_param_pos: int,
                             new_value: str,
                             controller: VistrailController) -> long
    Returns the id of the new parameter.
    """
    if controller is None:
        controller = get_current_controller()
    module = controller.current_pipeline.modules[module_id]
    function = module.functions[function_pos]
    old_param_id = function.params[old_param_pos].real_id
    controller.update_parameter(function, old_param_id, new_value)
    controller.updatePipelineScene()
    return function.params[old_param_pos].real_id

def add_port_spec(module_id, port_spec, controller=None):
    if controller is None:
        controller = get_current_controller()
    # module = controller.current_pipeline.modules[module_id]
    controller.add_module_port(module_id, (port_spec.type, port_spec.name,
                                           port_spec.sigstring))
    controller.updatePipelineScene()

##############################################################################

def select_version(version, ctrl=None):
    """select_version(int or str, ctrl=None):

    Given an integer, selects a version with the given number from the
    given vistrail (or the current one if no controller is given).

    Given a string, selects a version with that tag.

    """
    if ctrl is None:
        ctrl = get_current_controller()
    vistrail = ctrl.vistrail
    if isinstance(version, str):
        version = vistrail.get_tag_str(version).action_id
    ctrl.change_selected_version(version)
    ctrl.invalidate_version_tree(False)

def undo():
    get_current_vistrail_view().undo()

def redo():
    get_current_vistrail_view().redo()

def get_available_versions():
    """get_available_version(): ([int], {int: str})

    From the currently selected vistrail, return all available
    versions and the existing tags.

    """
    ctrl = get_current_controller()
    vistrail = ctrl.vistrail
    return (vistrail.actionMap.keys(), vistrail.get_tagMap())

def open_vistrail_from_file(filename):
    from vistrails.core.db.locator import FileLocator

    f = FileLocator(filename)
    view = get_builder_window().open_vistrail(f)
    return view

def close_vistrail(view, quiet=True):
    """close_vistrail(view: QVistrailView, quiet:bool)-> None
    Closes vistrail in view. If quiet is True it will discard changes
    automatically.
    
    """
    get_builder_window().close_vistrail(view, quiet=quiet)

def close_all_vistrails(quiet=True):
    get_builder_window().close_all_vistrails(quiet)
    
def new_vistrail():
    # Returns VistrailView - remember to be consistent about it..
    get_vistrails_application().builderWindow.new_vistrail(False)
    result = get_vistrails_application().builderWindow.get_current_view()
    return result

def get_vistrail_from_file(filename):
    from vistrails.core.db.locator import FileLocator
    from vistrails.core.vistrail.vistrail import Vistrail
    v = FileLocator(filename).load()
    if not isinstance(v, Vistrail):
        v = v.vistrail
    return v

##############################################################################
# Testing


class TestAPI(vistrails.gui.utils.TestVisTrailsGUI):

    def setUp(self):
        app = vistrails.gui.application.get_vistrails_application()
        app.builderWindow.auto_view = False
        app.builderWindow.close_all_vistrails(True)

    def test_close_current_vistrail_no_vistrail(self):
        self.assertRaises(NoVistrail, lambda: get_current_vistrail_view())

    def test_new_vistrail_no_save(self):
        v = new_vistrail()
        import vistrails.gui.vistrail_view
        assert isinstance(v, vistrails.gui.vistrail_view.QVistrailView)
        assert not v.controller.changed
        close_vistrail(v)

    def test_new_vistrail_button_states(self):
        assert get_vistrails_application().builderWindow.qactions['newVistrail'].isEnabled()
        assert not get_vistrails_application().builderWindow.qactions['closeVistrail'].isEnabled()
        assert not get_vistrails_application().builderWindow.qactions['saveFile'].isEnabled()
        assert not get_vistrails_application().builderWindow.qactions['saveFileAs'].isEnabled()
        view = new_vistrail()
        assert get_vistrails_application().builderWindow.qactions['newVistrail'].isEnabled()
        assert get_vistrails_application().builderWindow.qactions['closeVistrail'].isEnabled()
        self.assertEqual(get_vistrails_application().builderWindow.qactions['saveFile'].isEnabled(),
                         view.has_changes())
        assert get_vistrails_application().builderWindow.qactions['saveFileAs'].isEnabled()

    def test_detach_vistrail(self):
        view = new_vistrail()
        get_vistrails_application().builderWindow.detach_view(view)
        get_vistrails_application().builderWindow.attach_view(view)
        close_vistrail(view)

    
    
    



from vistrails.core.modules.vistrails_module import Module
from vistrails.core.modules.config import IPort, OPort
from vistrails.core.upgradeworkflow import  UpgradeModuleRemap

class TestUpgradeA(Module):
    _input_ports = [IPort("aaa", "basic:String")]
    _output_ports = [OPort("zzz", "basic:Integer")]

class TestUpgradeB(Module):
    _input_ports = [IPort("b", "basic:Integer")]

_modules = [TestUpgradeA, TestUpgradeB]
_upgrades = {"TestUpgradeA": 
             [UpgradeModuleRemap('0.8', '0.9', '0.9', None,
                                 function_remap={'a': 'aa'},
                                 src_port_remap={'z': 'zz'}),
              UpgradeModuleRemap('0.9', '1.0', '1.0', None,
                                 function_remap={'aa': 'aaa'},
                                 src_port_remap={'zz': 'zzz'})]}



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy

from vistrails.core.log.workflow_exec import WorkflowExec
from vistrails.db.domain import DBLog

class Log(DBLog):
    """ Class that stores info for logging a workflow execution. """

    def __init__(self, *args, **kwargs):
        DBLog.__init__(self, *args, **kwargs)

    def __copy__(self):
        return self.do_copy()

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBLog.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = Log
        return cp

    @staticmethod
    def convert(_log):
        if _log.__class__ == Log:
            return
        _log.__class__ = Log
        for workflow_exec in _log.workflow_execs:
            WorkflowExec.convert(workflow_exec)


    ##########################################################################
    # Properties

    def _get_id(self):
        return self.db_id
    def _set_id(self, id):
        self.db_id = id
    id = property(_get_id, _set_id)

    def _get_workflow_execs(self):
        return self.db_workflow_execs
    workflow_execs = property(_get_workflow_execs)
    def add_workflow_exec(self, wf_exec):
        self.db_add_workflow_exec(wf_exec)

    def delete_all_workflow_execs(self):
        for wf_exec in copy.copy(self.workflow_execs):
            self.db_delete_workflow_exec(wf_exec)

    def _get_vistrail_id(self):
        return self.db_vistrail_id
    def _set_vistrail_id(self, id):
        self.db_vistrail_id = id
    vistrail_id = property(_get_vistrail_id, _set_vistrail_id)

    def get_last_workflow_exec_id(self):
        if len(self.workflow_execs) < 1:
            return -1
        return max(wf_exec.id for wf_exec in self.workflow_execs)



#!/usr/bin/env python

# This example demonstrates the use and manipulation of lookup tables.

import vtk
from vtk.util.colors import *
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# First create pipeline a simple pipeline that reads a structure grid
# and then extracts a plane from the grid. The plane will be colored
# differently by using different lookup tables.
#
# Note: the Update method is manually invoked because it causes the
# reader to read; later on we use the output of the reader to set
# a range for the scalar values.
pl3d = vtk.vtkPLOT3DReader()
pl3d.SetXYZFileName(VTK_DATA_ROOT + "/Data/combxyz.bin")
pl3d.SetQFileName(VTK_DATA_ROOT + "/Data/combq.bin")
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.Update()
plane = vtk.vtkStructuredGridGeometryFilter()
plane.SetInputConnection(pl3d.GetOutputPort())
plane.SetExtent(1, 100, 1, 100, 7, 7)
lut = vtk.vtkLookupTable()
planeMapper = vtk.vtkPolyDataMapper()
planeMapper.SetLookupTable(lut)
planeMapper.SetInputConnection(plane.GetOutputPort())
planeMapper.SetScalarRange(pl3d.GetOutput().GetScalarRange())
planeActor = vtk.vtkActor()
planeActor.SetMapper(planeMapper)

# This creates an outline around the data.
outline = vtk.vtkStructuredGridOutlineFilter()
outline.SetInputConnection(pl3d.GetOutputPort())
outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outline.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)

# Much of the following is commented out. To try different lookup tables,
# uncommented the appropriate portions.

# This creates a black to white lut.
##lut.SetHueRange(0, 0)
##lut.SetSaturationRange(0, 0)
##lut.SetValueRange(0.2, 1.0)

# This creates a red to blue lut.
##lut.SetHueRange(0.0, 0.667)

# This creates a blue to red lut.
##lut.SetHueRange(0.667, 0.0)

# This creates a wierd effect. The Build() method causes the lookup
# table to allocate memory and create a table based on the currect
# hue, saturation, value, and alpha (transparency) range. Here we then
# manually overwrite the values generated by the Build() method.
lut.SetNumberOfColors(256)
lut.Build()
for i in range(0, 16):
    lut.SetTableValue(i*16, red[0], red[1], red[2], 1)
    lut.SetTableValue(i*16+1, green[0], green[1], green[2], 1)
    lut.SetTableValue(i*16+2, blue[0], blue[1], blue[2], 1)
    lut.SetTableValue(i*16+3, black[0], black[1], black[2], 1)


# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(outlineActor)
ren.AddActor(planeActor)

ren.SetBackground(0.1, 0.2, 0.4)
ren.TwoSidedLightingOff()

renWin.SetSize(250, 250)

cam1 = ren.GetActiveCamera()
cam1.SetClippingRange(3.95297, 50)
cam1.SetFocalPoint(8.88908, 0.595038, 29.3342)
cam1.SetPosition(-12.3332, 31.7479, 41.2387)
cam1.SetViewUp(0.060772, -0.319905, 0.945498)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""Testing package for console_mode"""

##############################################################################
import vistrails.core.modules
import vistrails.core.modules.module_registry
from vistrails.core.modules.vistrails_module import Module, ModuleError, NotCacheable
from vistrails.core.modules.basic_modules import Float, Integer

identifier = 'org.vistrails.vistrails.console_mode_test'
version = '0.9.1'
name = 'Console Mode Tests'
old_identifiers = ['edu.utah.sci.vistrails.console_mode_test']

class TestTupleExecution(Module):

    def compute(self):
        v1, v2 = self.get_input('input')
        self.set_output('output', v1 + v2)


class TestDynamicModuleError(Module):

    def compute(self):
        c = TestDynamicModuleError()
        c.die()

    def die(self):
        raise ModuleError(self, "I died!")

class TestChangeVistrail(NotCacheable, Module):

    def compute(self):
        if self.has_input('foo'):
            v1 = self.get_input('foo')
        else:
            v1 = 0
        if v1 != 12:
            self.change_parameter('foo', v1 + 1)

class TestCustomNamed(Module):

    pass


class TestOptionalPorts(Module):

    pass

##############################################################################

def initialize():
    reg = vistrails.core.modules.module_registry.get_module_registry()
    reg.add_module(TestTupleExecution)
    reg.add_input_port(TestTupleExecution, 'input', [Float, Float])
    reg.add_output_port(TestTupleExecution, 'output', (Float, 'output'))
    reg.add_module(TestDynamicModuleError)
    reg.add_module(TestChangeVistrail)
    reg.add_input_port(TestChangeVistrail, 'foo', Integer)

    reg.add_module(TestCustomNamed, name='different name')
    reg.add_input_port(TestCustomNamed, 'input', Float)

    reg.add_module(TestOptionalPorts)
    reg.add_input_port(TestOptionalPorts, 'foo', Float, optional=True)
    reg.add_output_port(TestOptionalPorts, 'foo', Float, optional=True)




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.system import _defaultPkgPrefix, \
    get_vistrails_basic_pkg_id, get_module_registry

def load_cls(cls_item, prefix=None):
    path = None
    if isinstance(cls_item, basestring):
        (path, cls_name) = cls_item.split(':')[:2]
    elif isinstance(cls_item, tuple):
        (path, cls_name) = cls_item
    if path is not None:
        try:
            module = __import__(path, globals(), locals(), [cls_name])
        except ImportError:
            if prefix is None:
                raise
            path = '.'.join([prefix, path])
            module = __import__(path, globals(), locals(), [cls_name])
        return getattr(module, cls_name)
    return cls_item

def create_descriptor_string(package, name, namespace=None,
                             use_package=True):
    package_str = ""
    namespace_str = ""
    if use_package:
        package_str = "%s:" % package
    if namespace:
        namespace_str = "%s|" % namespace
    return "%s%s%s" % (package_str, namespace_str, name)

def parse_descriptor_string(d_string, cur_package=None):
    """parse_descriptor_string will expand names of modules using
    information about the current package and allowing shortcuts
    for any bundled vistrails packages (e.g. "basic" for
    "org.vistrails.vistrails.basic").  It also allows a nicer
    format for namespace/module specification (namespace comes
    fist unlike port specifications where it is after the module
    name...

    Examples:
      "persistence:PersistentInputFile", None -> 
          ("org.vistrails.vistrails.persistence", PersistentInputFile", "")
      "basic:String", None ->
          ("org.vistrails.vistrails.basic", "String", "")
      "NamespaceA|NamespaceB|Module", "org.example.my" ->
          ("org.example.my", "Module", "NamespaceA|NamespaceB")
    """

    package = ''
    qual_name = ''
    name = ''
    namespace = None
    parts = d_string.strip().split(':', 1)
    if len(parts) > 1:
        qual_name = parts[1]
        if '.' in parts[0]:
            package = parts[0]
        else:
            package = '%s.%s' % (_defaultPkgPrefix, parts[0])
    else:
        qual_name = d_string
        if cur_package is None:
            reg = get_module_registry()
            if reg._current_package is not None:
                package = reg._current_package.identifier
            else:
                package = get_vistrails_basic_pkg_id()
        else:
            package = cur_package
    qual_parts = qual_name.rsplit('|', 1)
    if len(qual_parts) > 1:
        namespace, name = qual_parts
    else:
        name = qual_name
    return (package, name, namespace)

def parse_port_spec_item_string(spec, cur_package=None):
    spec = spec.strip()
    spec_arr = spec.split(':', 2)
    if len(spec_arr) > 2:
        # switch format of spec to more natural
        # <package>:<namespace>|<name> for descriptor parsing
        spec = '%s:%s|%s' % (spec_arr[0], spec_arr[2], spec_arr[1])
    return parse_descriptor_string(spec, cur_package)

def create_port_spec_item_string(package, name, namespace=None, 
                                 old_style=False):
    if old_style:
        if namespace:
            namespace = ':' + namespace
        return '%s:%s%s' % (package, name, namespace)
    else:
        return create_descriptor_string(package, name, namespace)

def parse_port_spec_string(p_string, cur_package=None):
    port_spec = p_string.strip()
    if port_spec.startswith('('):
        port_spec = port_spec[1:]
    if port_spec.endswith(')'):
        port_spec = port_spec[:-1]
    if port_spec.strip() == '':
        return []

    specs_list = []
    for spec in port_spec.split(','):
        specs_list.append(parse_port_spec_item_string(spec, 
                                                      cur_package))
    return specs_list


def create_port_spec_string(specs_list, old_style=False):
    spec_items = []
    for specs in specs_list:
        if len(specs) == 3:
            pkg, name, ns = specs
        elif len(specs) == 2:
            pkg, name = specs
            ns = None
        else:
            raise TypeError("create_port_spec_string() got spec tuple "
                            "with %d elements" % len(specs))
        spec_items.append(create_port_spec_item_string(pkg, name, ns,
                                                       old_style))
    return '(%s)' % ','.join(spec_items)

def expand_port_spec_string(p_string, cur_package=None, 
                            old_style=False):
    specs_list = parse_port_spec_string(p_string, cur_package)
    return create_port_spec_string(specs_list, old_style)


def make_modules_dict(*dcts, **kwargs):
    """Makes a dictionary suitable to be exposed as the '_modules' of a package.

    This combines dictionaries hierarchically to make a global dict from other
    "sub-dicts", allowing to structure a package with subpackages.

    For example, you could have:
        a/aa.py:
            _modules = {'aa': [Some, Module]}
        a/__init__.py:
            from .aa import _modules as aa_modules
            from .ab import _modules as bb_modules
            _modules = make_modules_dict(aa_modules, bb_modules, namespace='a')
        b.py:
            _modules = {'b': [Other, Modules]}
        init.py:
            from .a import _modules as a_modules
            from .b import _modules as b_modules
            _modules = make_modules_dict(a_modules, b_modules)

    The resulting module hierarchy will be:
        a.aa.Some, a.aa.Module
        b.Other, b.Modules
    """
    namespace = kwargs.pop('namespace', '')
    if kwargs:
        raise TypeError("make_modules_dict got unexpected keyword arguments")
    if namespace:
        build_namespace = lambda n: '%s|%s' % (namespace, n)
    else:
        build_namespace = lambda n: n

    dct = dict()
    for d in dcts:
        if not isinstance(d, dict):
            d = {'': d}

        for subname, sublist in d.iteritems():
            if subname:
                name = build_namespace(subname)
            else:
                name = namespace
            dct.setdefault(name, []).extend(sublist)

    return dct


###############################################################################

import unittest


class UnorderedList(object):
    def __init__(self, *elems):
        self._set = set(elems)

    def __eq__(self, other):
        return set(other) == self._set

    def __str__(self):
        return 'ul(%s)' % ', '.join(repr(e) for e in self._set)

    def __repr__(self):
        return str(self)


class TestModulesDict(unittest.TestCase):
    def test_modulesdict(self):
        ul = UnorderedList
        pkg_a_aa = {'aa': [1, 2], 'aa|n': [3]}
        pkg_a_ab = {'ab': [4, 5]}
        pkg_a = make_modules_dict(pkg_a_aa, pkg_a_ab, [12, 13], namespace='a')
        self.assertEqual(set(pkg_a.keys()),
                         set(['a', 'a|aa', 'a|aa|n', 'a|ab']))
        pkg_b = {'b': [6, 7]}
        glob = [10, 11]
        more = {'a|aa': [8, 9]}
        pkg = make_modules_dict(pkg_a, pkg_b, glob, more)
        self.assertEqual(
                pkg,
                {
                    '': ul(10, 11),
                    'a': ul(12, 13),
                    'a|aa': ul(1, 2, 8, 9),
                    'a|aa|n': ul(3),
                    'a|ab': ul(4, 5),
                    'b': ul(6, 7)})



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
# Author: Prabhu Ramachandran
# Copyright (c) 2004, Enthought, Inc.
# License: BSD Style.

"""This module generates the class hierarchy for any given Python
modules.  This can be used (among other things) to generate the
traitified VTK classes in the correct order.

"""

from __future__ import division

import __builtin__


class TreeNode(object):
    """Represents a node in the class tree.

    Stores information on the sub and super classes of a particular
    class.  It also stores the inheritance level of the class inside
    the inheritance tree (essentially how many levels of inheritance
    are there below this class).  This inheritance level is computed
    when the `get_level` method is called.  The `get_level` method
    works only when the parent information is complete.
    
    """

    def __init__(self, klass):
        """Given a class, create a node in the tree.

        Parameters
        ----------
        - klass : `class`

          The class which is represented as a node in the tree.

        """
        self.klass = klass
        self.name = klass.__name__
        self.children = []
        self.parents = []
        # Uninitialized level is set to None
        self.level = None

    def add_parent(self, parent):
        """Add a parent node."""
        assert isinstance(parent, TreeNode)
        if parent not in self.parents:
            self.parents.append(parent)

    def add_child(self, child):
        """Add a child node. """
        assert isinstance(child, TreeNode)
        if child not in self.children:
            self.children.append(child)

    def get_level(self):
        """Returns the inheritance level of the node (an int).  If the
        level has not been set, the method computes it.  Note however,
        that this computation will fail if the parent information is
        incorrect.

        """
        if not self.level:
            if self.parents:
                self.level = max([x.get_level() for x in self.parents]) + 1
            else:
                self.level = 0
        return self.level

    def get_ancestors(self):
        """Returns a list of ancestor nodes from which this class has
        descended.

        """
        def _get_ancestors(node, ancestors):
            ancestors.extend(node.parents)
            for p in node.parents:
                _get_ancestors(p, ancestors)
        ancestors = []
        _get_ancestors(self, ancestors)
        return ancestors


class ClassTree(object):
    """Contains and generates all the class tree information.

    On initialization of the instance, nothing is done.  The classes
    are obtained using the list of modules (or a single module) that
    is used to initialize the instance.  One must then call the
    `create` method to generate the tree structure.  The instance of
    the class also can be treated as an iterator which iterates over
    the nodes of the tree.

    There are two ways in which the tree hierarchy is stored.  A
    dictionary mapping class names to the tree node and a tree
    represented as a list of lists containing the nodes.  The tree is
    organized based on a concept of an inheritance level.  A class
    that has no parent classes (no base classes) is said to be at
    level zero.  If a class inherits successively from 7 classes, it
    is at level 6.  An example of inheritance for a vtkFoo class is
    given below:

      vtkFoo -> vtkBar -> vtkObject -> vtkObjectBase

    Here, vtkObjectBase has an inheritance level of 0 and vtkFoo a
    level of 3.  One can traverse the tree by using the level as an
    index and find all the classes at a particular level.

    Here is some example usage of this class::

        >>> import vtk
        >>> t = ClassTree(vtk)
        >>> t.create()
        >>> print t.get_node('vtkObject').name
        vtkObject
        >>> print t.get_node('vtkObject').parents[0].name
        vtkObjectBase
        >>> print len(t.tree[0])
        1
        >>> t.tree[0][0].name
        vtkObjectBase
    
    """

    def __init__(self, modules):
        """Initialize the instance with the given modules.

        Parameters
        ----------

        - modules : `sequence` of modules or a module

          This is either a single module or a sequence of modules.
          The instance uses these list of modules to generate the
          class tree.
          
        """
        self.nodes = {}
        self.tree = [[]]
        if not hasattr(modules, '__iter__'):
            self.modules = [modules]
        else:
            self.modules = modules

    def __iter__(self):
        return iter(self.nodes.values())

    def _generate_hierarchy(self, klass):
        """Does the hard work of generating the class hierarchy."""
        node = self.get_node(klass.__name__, create=1)
        for base in klass.__bases__:
            base_node = self.get_node_from_class(base, create=1)
            node.add_parent(base_node)
            base_node.add_child(node)
            self._generate_hierarchy(base)

    def get_class(self, name):
        """Given a class name in the given modules returns the class."""
        klass = None
        for m in self.modules:
            if hasattr(m, name):
                return getattr(m, name)
        if hasattr(__builtin__, name):
            klass = getattr(__builtin__, name)
        if not klass:
            try:
                klass = self.nodes[name].klass
            except KeyError:
                raise KeyError, "Cannot find class of name %s"%name
        return klass

    def add_node(self, klass):
        """Create a node for the given class."""
        name = klass.__name__
        if not self.nodes.has_key(name):
            node = TreeNode(klass)
            self.nodes[name] = node
            return node

    def get_node(self, name, create=0):
        """Get a node of the given name.

        Parameters
        ----------

        - name : `str`

          Name of the node to get.

        - create : `boolean`

          If True, a new node will be added if no node of the given
          name is available.  Defaults to False.

        Returns
        -------

        - `TreeNode`

        """
        if self.nodes.has_key(name):
            return self.nodes[name]
        elif create:
            return self.add_node(self.get_class(name))

    def get_node_from_class(self, cls, create=0):
        """Get a node of the given class.

        Parameters
        ----------

        - cls : `class`

          Class of the node to get.

        - create : `boolean`

          If True, a new node will be added if no node of the given
          name is available.  Defaults to False.

        Returns
        -------

        - `TreeNode`

        """
        name = cls.__name__
        if self.nodes.has_key(name):
            return self.nodes[name]
        elif create:
            return self.add_node(cls)

    def create(self, class_names=None):
        """This method generates the class tree given an optional list
        of class names.

        Parameters
        ----------

        - class_names - `list` of `str`

          An optional list of names of the classes to generate the
          tree for.  Defaults to None where the class list is computed
          from the modules.        

        """
        if class_names is None:
            class_names = []
            for m in self.modules:            
                class_names.extend(dir(m))
            
        # Generate the nodes.
        for name in class_names:
            klass = self.get_class(name)
            if klass and hasattr(klass, '__bases__'):
                self._generate_hierarchy(klass)

        # Compute the inheritance level and store the nodes in the tree.
        for node in self:
            d = node.get_level()
            while len(self.tree) <= d:
                self.tree.append([])
            self.tree[d].append(node)

        # Sort the nodes alphabetically.
        def _comp(x, y):
            return cmp(x.name, y.name)
        for nodes in self.tree:
            nodes.sort(_comp)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""
Dialog for web repository options
Includes login and upload tabs
"""
from __future__ import division

from PyQt4 import QtGui, QtCore
from vistrails.core.configuration import get_vistrails_configuration, get_vistrails_persistent_configuration
from vistrails.core.repository.poster.encode import multipart_encode
from vistrails.core.repository.poster.streaminghttp import register_openers
from vistrails.core.vistrail.controller import VistrailController
from vistrails.core.db.locator import ZIPFileLocator, FileLocator
from vistrails.core.db.io import load_vistrail
from vistrails.core import debug
import urllib, urllib2, cookielib
import os
import tempfile
import vistrails.api
import json

##############################################################################

class QRepositoryPushWidget(QtGui.QWidget):
    """ Tab that shows main repository options
        Allows users to login and push VisTrails to the Repository """

    def __init__(self, parent, status_bar, dialog):
        QtGui.QWidget.__init__(self, parent)
        self._status_bar = status_bar
        self.dialog = dialog

        base_layout = QtGui.QVBoxLayout(self)

        top = QtGui.QFrame(self)
        bottom = QtGui.QFrame(self)

        base_layout.addWidget(top)
        base_layout.addWidget(bottom, 1)

        # TODO: this '/' check should probably be done in core/configuration.py
        self.config = get_vistrails_configuration()
        if self.config.webRepositoryURL[-1] == '/':
            self.config.webRepositoryURL = self.config.webRepositoryURL[:-1]

        # check if the web repository url has changed while logged in
        if self.dialog.cookie_url and \
           self.dialog.cookie_url != self.config.webRepositoryURL:
            self.dialog.cookiejar = None



        ######################################################################
        # Detail Table
        bottom_layout = QtGui.QVBoxLayout(bottom)
        bottom_layout.setMargin(2)
        bottom_layout.setSpacing(2)

        # Show what workflows are unrunnble on the repository
        # and for what reasons
        self._unrunnable_table = QtGui.QTableWidget(0, 2, top)
        self._unrunnable_table.horizontalHeader().setStretchLastSection(True)

        bottom_layout.addWidget(self._unrunnable_table)

        self._repository_status = {}

        ######################################################################
        # Push info
        top_layout = QtGui.QVBoxLayout(top)

        self._vistrail_status_label = QtGui.QLabel("")
        self._vistrail_status_label.setWordWrap(True)
        top_layout.addWidget(self._vistrail_status_label)

        self.serverCombo = QtGui.QComboBox()
        self.connect(self.serverCombo,
                     QtCore.SIGNAL("currentIndexChanged(int)"),
                     self.check_dependencies)
        top_layout.addWidget(self.serverCombo)

        """
        self._default_perm_label = QtGui.QLabel("Default Global Permissions:")
        top_layout.addWidget(self._default_perm_label)
        self.perm_view = QtGui.QCheckBox("view")
        self.perm_download = QtGui.QCheckBox("download")
        self.perm_edit = QtGui.QCheckBox("edit")
        top_layout.addWidget(self.perm_view)
        top_layout.addWidget(self.perm_download)
        top_layout.addWidget(self.perm_edit)
        self.perm_view.setEnabled(True)
        """

        self.permission_gb = QtGui.QGroupBox(self)
        self.permission_gb.setTitle("Default Global Permissions")
        glayout = QtGui.QHBoxLayout()
        self.perm_view = QtGui.QCheckBox("view")
        self.perm_download = QtGui.QCheckBox("download")
        self.perm_edit = QtGui.QCheckBox("edit")
        glayout.addWidget(self.perm_view)
        glayout.addWidget(self.perm_download)
        glayout.addWidget(self.perm_edit)
        self.perm_view.setChecked(True)
        self.perm_download.setChecked(True)
        self.permission_gb.setLayout(glayout)
        top_layout.addWidget(self.permission_gb)

        self._details_label = QtGui.QLabel("")
        self._details_label.setWordWrap(True)
        top_layout.addWidget(self._details_label)

        for lbl in [self._details_label, self._vistrail_status_label]:
            lbl.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignLeft)
            lbl.setWordWrap(True)

        self._push_button = QtGui.QPushButton("&Push")
        self._push_button.setEnabled(False)
        self.connect(self._push_button,
                     QtCore.SIGNAL("clicked()"),
                     self.push_vistrail_to_repository)
        self._branch_button = QtGui.QPushButton("&Branch")
        self._branch_button.hide()
        self.connect(self._branch_button,
                     QtCore.SIGNAL("clicked()"),
                     (lambda branching=True : self.push_vistrail_to_repository(branching)))
        button_box = QtGui.QDialogButtonBox()
        button_box.addButton(self._push_button,
                             QtGui.QDialogButtonBox.ActionRole)
        button_box.addButton(self._branch_button,
                             QtGui.QDialogButtonBox.ActionRole)
        bottom_layout.addWidget(button_box)

    def populate_table(self):
        self._unrunnable_table.clear()

        # set horizontal headers
        wf_title = QtGui.QTableWidgetItem('Workflow')
        wf_title.setTextAlignment(QtCore.Qt.AlignCenter)
        self._unrunnable_table.setHorizontalHeaderItem(0, wf_title)

        detail_title = QtGui.QTableWidgetItem('Unsupported Modules/Packages')
        detail_title.setTextAlignment(QtCore.Qt.AlignCenter)
        self._unrunnable_table.setHorizontalHeaderItem(1, detail_title)

        # populate table
        count = 0
        for wf in self._unrunnable_wfs.keys():
            details = \
                    QtGui.QTableWidgetItem(', '.join(self._unrunnable_wfs[wf]))
            details.setTextAlignment(QtCore.Qt.AlignCenter)

            wf_item = QtGui.QTableWidgetItem(str(wf))
            wf_item.setTextAlignment(QtCore.Qt.AlignCenter)

            if count >= self._unrunnable_table.rowCount():
                self._unrunnable_table.insertRow(count)
            self._unrunnable_table.setItem(count, 0, wf_item)
            self._unrunnable_table.setItem(count, 1, details)
            count += 1

        # delete vertical headers
        for i in range(self._unrunnable_table.rowCount()):
            self._unrunnable_table.setVerticalHeaderItem(i, QtGui.QTableWidgetItem())

    def update_push_information(self):
        """ display push information text in this widget """
        self._vistrail_status_label.setText(self._repository_status['support_status'])
        self._details_label.setText(self._repository_status['details'])


    ##########################################################################
    # Signal handling

    def check_user_projects(self):
        # are we logged in?
        if not self.dialog.cookiejar:
            self._repository_status['support_status'] = "Please login"
            self._repository_status['details'] = ""
            self.update_push_information()
            self._push_button.setEnabled(False)
            self._branch_button.setEnabled(False)
            self._branch_button.hide()
        else:
            self._branch_button.setEnabled(False)
            self._branch_button.hide()
            server_url = "%s/projects/get_user_projects/" % \
                    self.config.webRepositoryURL
            register_openers(cookiejar=self.dialog.cookiejar)
            try:
                request = urllib2.Request(server_url)
                get_servers = urllib2.urlopen(request)
            except urllib2.HTTPError, e:
                self._repository_status['support_status'] = ""
                self._repository_status['details'] = ""
                if e.code == 500:
                    self._repository_status['support_status'] = \
                            ("Error checking user projects (server side issues)")
                    debug.critical("Error checking user projects (server side issues)")
                else:
                    debug.critical("Exception checking user projects", e)

                self._push_button.setEnabled(False)
                self.update_push_information()
                return

            servers = json.loads(get_servers.read())
            if not self.serverCombo.count():
                self.serverCombo.addItem("Please Select a Project", 0)
                for prj, srvr in servers:
                    self.serverCombo.addItem("%s (%s)" % (prj, srvr), [prj, srvr])



    def check_dependencies(self, index):
        """
        determines if current VisTrail will be
        supported by the repository's VisTrail server
        """

        if not index: return

        self._unrunnable_wfs = {}

        # are we logged in?
        if not self.dialog.cookiejar:
            self._repository_status['support_status'] = "Please login"
            self._repository_status['details'] = ""
            self.update_push_information()
            self._push_button.setEnabled(False)
            self._branch_button.hide()
        else:
            self.repository_supports_vt = True
            # get packages supported by VisTrails repository
            server = self.serverCombo.itemData(index)[1]
            packages_url = "%s/packages/supported_packages/%s" % \
                    (self.config.webRepositoryURL, server)


            try:
                get_supported_packages = urllib2.urlopen(url=packages_url)
            except urllib2.HTTPError, e:
                self._repository_status['support_status'] = ""
                self._repository_status['details'] = ""
                if e.code == 500:
                    self._repository_status['support_status'] = \
                            ("Error when checking dependencies (server side issues)")
                    debug.critical("Error when checking dependencies (server side issues)")
                else:
                    debug.critical("Exception checking dependencies", e)

                self._push_button.setEnabled(False)
                self.update_push_information()
                self.populate_table()
                return
            server_packages = json.loads(get_supported_packages.read())

            # get local packages and local data modules
            local_packages = []

            self.local_data_modules = ['File', 'FileSink', 'Path']
            self.unavailable_data = set()
            self.unsupported_packages = set()
            has_python_source = False

            vistrail = vistrails.api.get_current_vistrail()

            for version_id in vistrail.get_tagMap():
                pipeline = vistrail.getPipeline(version_id)
                tag = vistrail.get_tagMap()[version_id]
                workflow_packages = []
                for module in pipeline.module_list:
                    # count modules that use data unavailable to web repository
                    on_repo = False
                    if module.name[-6:] == 'Reader' or \
                       module.name in self.local_data_modules:
                        for edge in pipeline.graph.edges_to(module.id):
                            if pipeline.modules[edge[0]].name in ['DownloadFile',
                                                                  'RepoSync']:
                                on_repo = True

                        if not on_repo:
                            if tag not in self._unrunnable_wfs.keys():
                                self._unrunnable_wfs[tag] = []
                            self._unrunnable_wfs[tag].append(module.name)
                            self.unavailable_data.add(module.name)

                    if module.name == "PythonSource":
                        if tag not in self._unrunnable_wfs.keys():
                            self._unrunnable_wfs[tag] = []
                        self._unrunnable_wfs[tag].append(module.name)
                        has_python_source = True

                    # get all packages used in tagged versions of this workflow
                    if module.package not in workflow_packages:
                        workflow_packages.append(module.package)

                # find unsupported packages
                wf_unsupported_packages = filter((lambda p: p not in \
                                                  server_packages),
                                                 workflow_packages)
                if wf_unsupported_packages:
                    if tag not in self._unrunnable_wfs.keys():
                        self._unrunnable_wfs[tag] = []
                    for package in wf_unsupported_packages:
                        self._unrunnable_wfs[tag].append(package)
                        self.unsupported_packages.add(package)

            # display unsupported packages
            self._repository_status['details'] = "Details:\n"
            if self.unsupported_packages:
                self.repository_supports_vt = False
                self._repository_status['details'] += \
                        "Packages incompatible with web repository: %s\n\n" % \
                        (', '.join(self.unsupported_packages)[:-2])

            # display unsupported data modules
            if len(self.unavailable_data):
                self.repository_supports_vt = False
                self._repository_status['details'] += \
                        "Data sources not available on web repository: %s\n"%\
                        (', '.join(self.unavailable_data)[:-2])

            if has_python_source:
                self.repository_supports_vt = False
                self._repository_status['details'] += \
                        ("This Vistrail contains PythonSource module(s) and "
                         "will have to be verified by admins before it can be "
                         "run on the web repository. You will be notified when"
                         " your workflows have been verfied")

            self._repository_status['support_status'] = ""
            controller = vistrails.api.get_current_controller()
            if controller.vistrail.get_annotation('repository_vt_id'):
                # Since repository_vt_id doesn't mirror crowdlabs vt id
                # get the crowdlabs vt id for linkage
                vt_url = "%s/vistrails/details/%s" % \
                        (self.config.webRepositoryURL,
                         controller.vistrail.get_annotation('repository_vt_id').value)
                try:
                    request = urllib2.urlopen(url=vt_url)
                    # TODO: check project that vistrail is part of and notify user
                    # that branching will add it to a different project
                    if request.code == 200:
                        # the vistrail exists on the repository, setup merge settings
                        self.permission_gb.setTitle(("Default Global Permissions "
                                                     "(only applicable to branching):"))
                        self._push_button.setText("Commit changes")
                        self._branch_button.setEnabled(True)
                        self._branch_button.show()

                        vistrail_link = "%s/vistrails/details/%s" % \
                                (self.config.webRepositoryURL,
                                 controller.vistrail.get_annotation('repository_vt_id').value)

                        self._repository_status['support_status'] = \
                                ("You are attempting to update this Vistrail: "
                                 "<a href='%s'>%s</a>. This will possibly update your local version with changes from the web repository<br><br>") % \
                                (vistrail_link, vistrail_link)

                except urllib2.HTTPError:
                    # the vistrail has *probably* been deleted or doesn't exist
                    # remove repository_vt_id annotation
                    repo_annotation = controller.vistrail.get_annotation('repository_vt_id')
                    if repo_annotation:
                        controller.vistrail.db_delete_annotation(repo_annotation)

            if self.repository_supports_vt:
                self._repository_status['support_status'] += \
                        ("All of this Vistrail's tagged versions are supported"
                         " by the Repository.")
            else:
                self._repository_status['support_status'] += \
                        ("This Vistrail contains packages or modules that are"
                         " not supported by the Repository.<br>"
                         "You may still upload the VisTrail but it will "
                         "not be run by the Repository.")


            self._push_button.setEnabled(True)

            self.update_push_information()
            self.populate_table()

    def push_vistrail_to_repository(self, branching=False):
        """ uploads current VisTrail to web repository """

        self._repository_status['details'] = "Pushing to repository..."
        self._push_button.setEnabled(False)
        self._branch_button.setEnabled(False)
        self.update_push_information()
        try:
            # create temp file
            (fd, filename) = tempfile.mkstemp(suffix='.vt', prefix='vt_tmp')
            os.close(fd)

            # writing tmp vt and switching back to original vt
            locator = ZIPFileLocator(filename)
            controller = vistrails.api.get_current_controller()
            controller.write_vistrail(locator, export=True)

            # check if this vt is from the repository
            if controller.vistrail.get_annotation('repository_vt_id'):
                repository_vt_id = controller.vistrail.get_annotation('repository_vt_id').value
            else:
                repository_vt_id = -1

            # upload vistrail temp file to repository
            register_openers(cookiejar=self.dialog.cookiejar)
            project = self.serverCombo.itemData(self.serverCombo.currentIndex())[0]
            if project == "Default": project = ""

            params = {'vistrail_file': open(filename, 'rb'),
                      'action': 'upload',
                      'name': controller.locator.short_name,
                      'repository_vt_id': repository_vt_id if not branching else -1,
                      'is_runnable': not bool(len(self.unsupported_packages)+ \
                                              len(self.local_data_modules)),
                      'vt_id': 0,
                      'branched_from': "" if not branching else repository_vt_id,
                      'project': project,
                      'everyone_can_view': self.perm_view.checkState(),
                      'everyone_can_edit': self.perm_edit.checkState(),
                      'everyone_can_download': self.perm_download.checkState()
                     }

            upload_url = "%s/vistrails/remote_upload/" % \
                    self.config.webRepositoryURL

            datagen, headers = multipart_encode(params)
            request = urllib2.Request(upload_url, datagen, headers)
            result = urllib2.urlopen(request)
            updated_response = result.read()

            os.unlink(filename)

            if updated_response[:6] == "upload":
                # No update, just upload
                if result.code != 200:
                    self._repository_status['details'] = \
                            "Push to repository failed"
                    debug.critical("Push to repository failed (Please contact an administrator)")
                else:
                    repository_vt_id = int(updated_response[8:])
                    controller.vistrail.set_annotation('repository_vt_id',
                                                       repository_vt_id)
                    controller.vistrail.set_annotation('repository_creator',
                                                       self.dialog.loginUser)
                    # ensure that the annotations get saved
                    controller.set_changed(True)
                    self._repository_status['details'] = \
                            "Push to repository was successful"
            else:
                # update, load updated vistrail
                if result.code != 200:
                    self._repository_status['details'] = "Update Failed"
                    debug.critical("Update vistrail in web repository failed (Please contact an administrator)")
                else:
                    debug.log("getting version from web")
                    # request file to download
                    download_url = "%s/vistrails/download/%s/" % \
                            (self.config.webRepositoryURL, updated_response)

                    request = urllib2.Request(download_url)
                    result = urllib2.urlopen(request)
                    updated_file = result.read()

                    # create temp file of updated vistrail
                    (fd, updated_filename) = tempfile.mkstemp(suffix='.vtl',
                                                              prefix='vtl_tmp')
                    os.close(fd)
                    updated_vt = open(updated_filename, 'w')
                    updated_vt.write(updated_file)
                    updated_vt.close()

                    # switch vistrails to updated one
                    controller = vistrails.api.get_current_controller()

                    updated_locator = FileLocator(updated_filename)

                    (up_vistrail, abstractions, thumbnails, mashups) = \
                            load_vistrail(updated_locator)

                    # FIXME need to figure out what to do with this !!!
                    current_version = controller.current_version
                    controller.set_vistrail(up_vistrail,
                                            controller.vistrail.locator,
                                            abstractions, thumbnails, mashups)
                    controller.change_selected_version(current_version)
                    # update version tree drawing
                    controller.recompute_terse_graph()
                    controller.invalidate_version_tree()

                    os.remove(updated_filename)
                    os.remove(updated_filename[:-1])

                    self._repository_status['details'] = \
                            "Update to repository was successful"

        except Exception, e:
            debug.critical("An error occurred", e)
            self._repository_status['details'] = "An error occurred"
        self.update_push_information()

class QRepositoryLoginPopup(QtGui.QDialog):
    """ Dialog that shows repository authentication """

    def __init__(self, parent, status_bar, dialog):
        QtGui.QDialog.__init__(self, parent)
        self._status_bar = status_bar
        self.dialog = dialog
        self.dialog.cookiejar = None

        self.config = get_vistrails_configuration()
        # TODO: this '/' check should probably be done in core/configuration.py
        if self.config.webRepositoryURL[-1] == '/':
            self.config.webRepositoryURL = self.config.webRepositoryURL[:-1]

        self.setWindowTitle('Log in to Web Repository')

        grid_layout = QtGui.QGridLayout(self)
        grid_layout.setSpacing(10)

        l1 = QtGui.QLabel("Repository location: %s" % \
                          self.config.webRepositoryURL, self)
        grid_layout.addWidget(l1, 0, 0, 1, 2)

        l2 = QtGui.QLabel("Username:", self)
        grid_layout.addWidget(l2, 1, 0)

        if self.config.check('webRepositoryUser'):
            self.loginUser = QtGui.QLineEdit(self.config.webRepositoryUser, self)
        else:
            self.loginUser = QtGui.QLineEdit("", self)

        self.loginUser.setFixedWidth(200)
        self.loginUser.setSizePolicy(QtGui.QSizePolicy.Fixed,
                                     QtGui.QSizePolicy.Fixed)
        grid_layout.addWidget(self.loginUser, 1, 1)

        l3 = QtGui.QLabel("Password:", self)
        grid_layout.addWidget(l3, 2, 0)

        self.loginPassword = QtGui.QLineEdit("", self)
        self.loginPassword.setEchoMode(2)
        self.loginPassword.setFixedWidth(200)
        self.loginPassword.setSizePolicy(QtGui.QSizePolicy.Fixed,
                                     QtGui.QSizePolicy.Fixed)
        grid_layout.addWidget(self.loginPassword, 2, 1)

        self.saveLogin = QtGui.QCheckBox("Save username", self)
        if self.config.check('webRepositoryUser'):
            self.saveLogin.setChecked(True)
        grid_layout.addWidget(self.saveLogin, 3, 0)

        self._login_status_label= QtGui.QLabel("", self)
        grid_layout.addWidget(self._login_status_label, 4, 0)

        self._login_button = QtGui.QPushButton("&Login", self)
        self._cancel_button = QtGui.QPushButton("&Cancel", self)

        self.connect(self._login_button,
                     QtCore.SIGNAL("clicked()"),
                     self.clicked_on_login)

        self.connect(self._cancel_button,
                     QtCore.SIGNAL("clicked()"),
                     self.clicked_on_cancel)

        button_box = QtGui.QDialogButtonBox()
        button_box.addButton(self._login_button,
                             QtGui.QDialogButtonBox.ActionRole)
        button_box.addButton(self._cancel_button,
                             QtGui.QDialogButtonBox.ActionRole)

        grid_layout.addWidget(button_box, 5, 0)

        for lbl in [l1, l2, l3, self.loginUser, self.loginPassword, self._login_status_label]:
            lbl.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignLeft)

        grid_layout.setRowStretch(1, 1)
        grid_layout.setColumnStretch(1, 1)

        self.loginUser.setEnabled(True)
        self.loginPassword.setEnabled(True)
        self._login_button.setEnabled(True)
        self.saveLogin.setEnabled(True)


        if self.dialog.cookiejar and \
           'sessionid' in [cookie.name for cookie in self.dialog.cookiejar]:
            self.loginUser.setEnabled(False)
            self.loginPassword.setEnabled(False)
            self._login_button.setEnabled(False)
            self.saveLogin.setEnabled(False)

    def show_login_information(self):
        """ display/update login info text """
        self._login_status_label.setText(self._login_status)


    ##########################################################################
    # Signal handling
    def clicked_on_login(self):
        """
        Attempts to log into web repository
        stores auth cookie for session
        """
        from vistrails.gui.application import get_vistrails_application

        self.dialog.loginUser = self.loginUser.text()
        params = urllib.urlencode({'username':self.dialog.loginUser,
                                   'password':self.loginPassword.text()})
        self.dialog.cookiejar = cookielib.CookieJar()

        # set base url used for cookie
        self.dialog.cookie_url = self.config.webRepositoryURL

        self.loginOpener = urllib2.build_opener(urllib2.HTTPCookieProcessor(self.dialog.cookiejar))

        # FIXME doesn't use https
        login_url = "%s/account/login/" % self.config.webRepositoryURL
        request = urllib2.Request(login_url, params)
        url = self.loginOpener.open(request)

        # login failed
        if not 'sessionid' in [cookie.name for cookie in self.dialog.cookiejar]:
            debug.critical("Incorrect username or password")
            self.dialog.cookiejar = None
        else: # login successful

            self._login_status = "login successful"

            self.loginUser.setEnabled(False)
            self.loginPassword.setEnabled(False)
            self._login_button.setEnabled(False)
            self.saveLogin.setEnabled(False)

            # add association between VisTrails user and web repository user
            if self.saveLogin.checkState():
                if not (self.config.check('webRepositoryUser') and self.config.webRepositoryUser == self.loginUser.text()):
                    self.config.webRepositoryUser = str(self.loginUser.text())
                    pers_config = get_vistrails_persistent_configuration()
                    pers_config.webRepositoryUser = self.config.webRepositoryUser
                    get_vistrails_application().save_configuration()

            # remove association between VisTrails user and web repository user
            else:
                if self.config.check('webRepositoryUser') and self.config.webRepositoryUser:
                    self.config.webRepositoryUser = ""
                    pers_config = get_vistrails_persistent_configuration()
                    pers_config.webRepositoryUser = ""
                    get_vistrails_application().save_configuration()
            self.close_dialog(0)

    def clicked_on_cancel(self):
        self.reject()
 
    def close_dialog(self, status=0):
        self.accept()

class QRepositoryDialog(QtGui.QDialog):
    """ Dialog that shows repository options """

    cookiejar = None
    cookie_url = None
    loginUser = None
    def __init__(self, parent):
        QtGui.QDialog.__init__(self, parent)
        self._status_bar = QtGui.QStatusBar(self)
        self.setWindowTitle('Push vistrail to Web Repository')

        l = QtGui.QVBoxLayout(self)
        l.setMargin(0)
        l.setSpacing(0)

        widget = QtGui.QWidget(self)
        user_layout = QtGui.QHBoxLayout(widget)
        self._user_text = QtGui.QLabel('N/A')
        user_layout.addWidget(self._user_text)
        self._logout_button = QtGui.QPushButton('Log out')
        #self._logout_button.setFixedWidth(100)
        user_layout.addWidget(self._logout_button)
        user_layout.addStretch(1)
        l.addWidget(widget)
        self._push_widget = QRepositoryPushWidget(self, self._status_bar,
                                                  self.__class__)
        l.addWidget(self._push_widget)

        self._button_box = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Close,
                                                  QtCore.Qt.Horizontal,
                                                  self)
        
        self.connect(self._button_box,
                     QtCore.SIGNAL('clicked(QAbstractButton *)'),
                     self.close_dialog)

        self.connect(self._logout_button,
                     QtCore.SIGNAL('clicked()'),
                     self.clicked_on_logout)

        l.addWidget(self._button_box)
        l.addWidget(self._status_bar)

        self.show_login_dialog()

        self._push_widget.check_user_projects()

    def clicked_on_logout(self):
        """ Reset cookie and show login dialog """
        self.cookiejar = None
        self.show_login_dialog()

    def show_login_dialog(self):
        """ Prompt for login """
        if not self.cookiejar:
            loginDialog = QRepositoryLoginPopup(self, self._status_bar, self.__class__)
            res = loginDialog.exec_()
            if res == QtGui.QDialog.Rejected:
                self.close_dialog()
                return
        self._user_text.setText('Logged in as: %s@%s' % (self.loginUser, self.cookie_url))
        self._push_widget.check_user_projects()

    def close_dialog(self):
        self.done(0)

    def tab_changed(self, index):
        """ tab_changed(index: int) -> None """
        if index == 1: # push tab
            self._push_tab.check_user_projects()

    def sizeHint(self):
        return QtCore.QSize(800, 600)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file contains widgets that can be used for dropping Constant class
variables. It will construct an input form for the value.

QVariableDropBox
QVerticalWidget
QVariableInputWidget
QVariableInputForm
QDragVariableLabel
QHoverVariableLabel
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core import debug
from vistrails.core.vistrail.module_function import ModuleFunction
from vistrails.core.vistrail.module_param import ModuleParam
from vistrails.core.modules import module_registry
from vistrails.core.modules.basic_modules import Constant
from vistrails.core.vistrail.vistrailvariable import VistrailVariable
from vistrails.gui.common_widgets import QPromptWidget
from vistrails.gui.modules.utils import get_widget_class
from vistrails.gui.modules.constant_configuration import StandardConstantWidget
from vistrails.gui.module_palette import QModuleTreeWidget
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.utils import show_question, YES_BUTTON, NO_BUTTON
import uuid

################################################################################

class QVariableDropBox(QtGui.QScrollArea):
    """
    QVariableDropBox is just a widget such that items that subclass
    Constant from the module palette can be dropped into its client rect.
    It then constructs an input form based on the type of handling widget
    
    """
    def __init__(self, parent=None):
        """ QVariableDropBox(parent: QWidget) -> QVariableDropBox
        Initialize widget constraints
        
        """
        QtGui.QScrollArea.__init__(self, parent)
        self.setAcceptDrops(True)
        self.setWidgetResizable(True)
        self.vWidget = QVerticalWidget()
        self.setWidget(self.vWidget)
        self.updateLocked = False
        self.controller = None

    def dragEnterEvent(self, event):
        """ dragEnterEvent(event: QDragEnterEvent) -> None
        Set to accept drops from the module palette
        
        """
        if isinstance(event.source(), QModuleTreeWidget):
            data = event.mimeData()
            if hasattr(data, 'items'):
                event.accept()
        else:
            event.ignore()
        
    def dragMoveEvent(self, event):
        """ dragMoveEvent(event: QDragMoveEvent) -> None
        Set to accept drag move event from the module palette
        
        """
        if isinstance(event.source(), QModuleTreeWidget):
            data = event.mimeData()
            if hasattr(data, 'items'):
                event.accept()

    def dropEvent(self, event):
        """ dropEvent(event: QDragMoveEvent) -> None
        Accept drop event to add a new variable
        
        """
        if isinstance(event.source(), QModuleTreeWidget):
            data = event.mimeData()
            if hasattr(data, 'items'):
                event.accept()
                assert len(data.items) == 1
                item = data.items[0]
                if issubclass(item.descriptor.module, Constant):
                    if item.descriptor and self.controller:
                        self.lockUpdate()
                        (text, ok) = QtGui.QInputDialog.getText(self,
                                                                'Set Variable Name',
                                                                'Enter the variable name',
                                                                QtGui.QLineEdit.Normal,
                                                                '')
                        var_name = str(text).strip()
                        while ok and self.controller.check_vistrail_variable(var_name):
                            msg =" This variable name is already being used.\
 Please enter a different variable name "
                            (text, ok) = QtGui.QInputDialog.getText(self,
                                                                    'Set Variable Name',
                                                                    msg,
                                                                    QtGui.QLineEdit.Normal,
                                                                    text)
                            var_name = str(text).strip()
                        if ok:
                            self.vWidget.addVariable(str(uuid.uuid1()), var_name, item.descriptor)
                            self.scrollContentsBy(0, self.viewport().height())
                        self.unlockUpdate()
                #self.emit(QtCore.SIGNAL("paramsAreaChanged"))

    def updateController(self, controller):
        """ updateController(controller: VistrailController) -> None
        Construct input forms for a controller's variables
        
        """
        # we shouldn't do this whenver the controller changes...
        if self.controller != controller:
            self.controller = controller
            if self.updateLocked: return
            self.vWidget.clear()
            if controller:
                reg = module_registry.get_module_registry()
                for var in [v for v in controller.vistrail.vistrail_vars]:
                    try:
                        descriptor = reg.get_descriptor_by_name(var.package,
                                                                var.module, 
                                                                var.namespace)
                    except module_registry.ModuleRegistryException:
                        debug.critical("Missing Module Descriptor for vistrail"
                                       " variable %s\nPackage: %s\nType: %s"
                                       "\nNamespace: %s" % \
                                           (var.name, var.package, var.module, 
                                            var.namespace))
                        continue
                    self.vWidget.addVariable(var.uuid, var.name, descriptor, 
                                             var.value)
                self.vWidget.showPromptByChildren()
            else:
                self.vWidget.showPrompt(False)

    def lockUpdate(self):
        """ lockUpdate() -> None
        Do not allow updateModule()
        
        """
        self.updateLocked = True
        
    def unlockUpdate(self):
        """ unlockUpdate() -> None
        Allow updateModule()
        
        """
        self.updateLocked = False

class QVerticalWidget(QPromptWidget):
    """
    QVerticalWidget is a widget holding other variable widgets
    vertically
    
    """
    def __init__(self, parent=None):
        """ QVerticalWidget(parent: QWidget) -> QVerticalWidget
        Initialize with a vertical layout
        
        """
        QPromptWidget.__init__(self, parent)
        self.setPromptText("Drag a constant from the Modules panel to create a variable")
        self.setLayout(QtGui.QVBoxLayout())
        self.layout().setMargin(0)
        self.layout().setSpacing(5)
        self.layout().setAlignment(QtCore.Qt.AlignTop)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding,
                           QtGui.QSizePolicy.Expanding)
        self.setMinimumHeight(20)
        self._variable_widgets = []
        
    def addVariable(self, uuid, name, descriptor, value=''):
        """ addVariable(uuid:str, name: str, descriptor: ModuleDescriptor, value: str) -> None
        Add an input form for the variable
        
        """
        inputForm = QVariableInputWidget(uuid, name, descriptor, value, self)
        self.connect(inputForm, QtCore.SIGNAL('deleted(QWidget*)'), 
                     self.delete_form)
        self.layout().addWidget(inputForm)
        inputForm.show()
        self.setMinimumHeight(self.layout().minimumSize().height())
        self.showPrompt(False)
        self._variable_widgets.append(inputForm)

    def clear(self):
        """ clear() -> None
        Clear and delete all widgets in the layout
        
        """
        self.setEnabled(False)
        for v in self._variable_widgets:
            self.disconnect(v, QtCore.SIGNAL('deleted(QWidget*)'), 
                         self.delete_form)
            self.layout().removeWidget(v)
            v.setParent(None)
            v.deleteLater()
        self._variable_widgets = []
        self.setEnabled(True)

    def delete_form(self, input_form):
        self.disconnect(input_form, QtCore.SIGNAL('deleted(QWidget*)'), 
                     self.delete_form)
        var_name = input_form.var_name
        variableBox = self.parent().parent()
        self.layout().removeWidget(input_form)
        self._variable_widgets.remove(input_form)
        input_form.setParent(None)
        input_form.deleteLater()
        self.showPromptByChildren()

        if variableBox.controller:
            variableBox.lockUpdate()
            variableBox.controller.set_vistrail_variable(var_name, None)
            variableBox.unlockUpdate()
        self.setMinimumHeight(self.layout().minimumSize().height())


class QVariableInputWidget(QtGui.QDockWidget):
    def __init__(self, uuid, name, descriptor, value='', parent=None):
        QtGui.QDockWidget.__init__(self, parent)
        self.var_uuid = uuid
        self.var_name = name
        self.descriptor = descriptor
        self.setFeatures(QtGui.QDockWidget.DockWidgetClosable)
        # Create group and titlebar widgets for input widget
        self.group_box = QVariableInputForm(descriptor, value, self)
        self.setWidget(self.group_box)
        title_widget = QtGui.QWidget()
        title_layout = QtGui.QHBoxLayout()
        self.closeButton = QtGui.QToolButton()
        self.closeButton.setAutoRaise(True)
        self.closeButton.setIcon(QtGui.QIcon(self.style().standardPixmap(QtGui.QStyle.SP_TitleBarCloseButton)))
        self.closeButton.setIconSize(QtCore.QSize(13, 13))
        self.closeButton.setFixedWidth(16)
        self.label = QHoverVariableLabel(name)
        title_layout.addWidget(self.label)
        title_layout.addWidget(self.closeButton)
        title_widget.setLayout(title_layout)
        self.setTitleBarWidget(title_widget)
        self.connect(self.closeButton, QtCore.SIGNAL('clicked()'), self.close)
        
    def renameVariable(self, var_name):
        # First delete old var entry
        variableBox = self.parent().parent().parent()
        if variableBox.controller:
            variableBox.lockUpdate()
            variableBox.controller.set_vistrail_variable(self.var_name, None, False)
            variableBox.unlockUpdate()
        # Create var entry with new name, but keeping the same uuid
        self.var_name = var_name
        self.label.setText(var_name)
        self.group_box.updateMethod()
        
    def closeEvent(self, event):
        choice = show_question('Delete %s?'%self.var_name,
           'Are you sure you want to permanently delete the VisTrail variable\
 "%s"?\n\nNote:  Any workflows using this variable will be left in an invalid state.'%self.var_name,
                               [NO_BUTTON,YES_BUTTON],
                               NO_BUTTON)
        if choice == NO_BUTTON:
            event.ignore()
            return
        self.emit(QtCore.SIGNAL('deleted(QWidget*)'), self)

    def keyPressEvent(self, e):
        if e.key() in [QtCore.Qt.Key_Delete, QtCore.Qt.Key_Backspace]:
            self.close()
        else:
            QtGui.QDockWidget.keyPressEvent(self, e)
    
    def check_variable(self, name):
        """ check_variable(name: str) -> Boolean
        Returns True if the vistrail already has the variable name

        """
        variableBox = self.parent().parent().parent()
        if variableBox.controller:
            return variableBox.controller.check_vistrail_variable(name)
        return False

class QVariableInputForm(QtGui.QGroupBox):
    """
    QVariableInputForm is a widget with multiple input lines depends on
    the method signature
    
    """
    def __init__(self, descriptor, var_strValue="", parent=None):
        """ QVariableInputForm(descriptor: ModuleDescriptor, var_strValue: str,
                               parent: QWidget) -> QVariableInputForm
        Initialize with a vertical layout
        
        """
        QtGui.QGroupBox.__init__(self, parent)
        self.setLayout(QtGui.QGridLayout())
        self.layout().setMargin(5)
        self.layout().setSpacing(5)
        self.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.setSizePolicy(QtGui.QSizePolicy.Preferred,
                           QtGui.QSizePolicy.Fixed)
        self.palette().setColor(QtGui.QPalette.Window,
                                CurrentTheme.METHOD_SELECT_COLOR)
        # Create widget for editing variable
        p = ModuleParam(type=descriptor.name, identifier=descriptor.identifier,
                        namespace=descriptor.namespace)
        p.strValue = var_strValue
        widget_type = get_widget_class(descriptor)
        self.widget = widget_type(p, self)
        self.label = QDragVariableLabel(p.type)
        self.layout().addWidget(self.label, 0, 0)
        self.layout().addWidget(self.widget, 0, 1)
        self.updateMethod()

    def focusInEvent(self, event):
        """ gotFocus() -> None
        Make sure the form painted as selected
        
        """
        self.setAutoFillBackground(True)

    def focusOutEvent(self, event):
        """ lostFocus() -> None
        Make sure the form painted as non-selected and then
        perform a parameter changes
        
        """
        self.setAutoFillBackground(False)

    def updateMethod(self):
        """ updateMethod() -> None
        Update the variable values in vistrail controller
        
        """
        inputWidget = self.parent()
        variableBox = inputWidget.parent().parent().parent()
        if variableBox.controller:
            variableBox.lockUpdate()
            descriptor = inputWidget.descriptor
            var = VistrailVariable(inputWidget.var_name, inputWidget.var_uuid,
                                   descriptor.identifier, descriptor.name,
                                   descriptor.namespace, str(self.widget.contents()))
            variableBox.controller.set_vistrail_variable(inputWidget.var_name, var)
            variableBox.unlockUpdate()

class QDragVariableLabel(QtGui.QLabel):
    """
    QDragVariableLabel is a QLabel that can be dragged to connect
    to an input port
    """
    def __init__(self, var_type='', parent=None):
        """ QDragVariableLabel(var_type:str,
                                parent: QWidget) -> QDragVariableLabel
        Initialize the label with a variable type
        
        """
        QtGui.QLabel.__init__(self, parent)
        self.var_type = var_type
        self.setText(var_type)
        self.setAttribute(QtCore.Qt.WA_Hover)
        self.setCursor(CurrentTheme.OPEN_HAND_CURSOR)
        self.setToolTip('Drag to an input port')
        self.palette().setColor(QtGui.QPalette.WindowText,
                                CurrentTheme.HOVER_DEFAULT_COLOR)

    def event(self, event):
        """ event(event: QEvent) -> Event Result
        Override to handle hover enter and leave events for hot links
        
        """
        if event.type()==QtCore.QEvent.HoverEnter:
            self.palette().setColor(QtGui.QPalette.WindowText,
                                    CurrentTheme.HOVER_SELECT_COLOR)
        if event.type()==QtCore.QEvent.HoverLeave:
            self.palette().setColor(QtGui.QPalette.WindowText,
                                    CurrentTheme.HOVER_DEFAULT_COLOR)
        return QtGui.QLabel.event(self, event)

    def mousePressEvent(self, event):
        """ mousePressEvent(event: QMouseEvent) -> None        
        If mouse click on the label, show up a dialog to change/add
        the variable name
        
        """
        if event.button()==QtCore.Qt.LeftButton:
            inputWidget = self.parent().parent()
            var_name = inputWidget.var_name
            var_uuid = inputWidget.var_uuid
            # Create pixmap from variable name and type
            drag_str = var_name + ' : ' + self.var_type
            drag_label = QDragVariableLabel(drag_str)
            drag_label.adjustSize()
            painter = QtGui.QPainter()
            font = QtGui.QFont()
            size = drag_label.size()
            image = QtGui.QImage(size.width()+4, size.height()+4, QtGui.QImage.Format_ARGB32_Premultiplied)
            image.fill(0)
            painter.begin(image)
            painter.setPen(QtCore.Qt.NoPen)
            painter.setBrush(self.palette().highlight())
            painter.drawRect(QtCore.QRectF(0, 0, image.width(), image.height()))
            painter.setFont(font)
            painter.setPen(QtCore.Qt.black)
            painter.drawText(QtCore.QRect(QtCore.QPoint(2,2), size), QtCore.Qt.AlignLeft | QtCore.Qt.TextSingleLine, drag_str)
            painter.end()
            pixmap = QtGui.QPixmap.fromImage(image)
            # Create drag action
            mimeData = QtCore.QMimeData()
            portspec = inputWidget.descriptor.get_port_spec('value', 'output')
            mimeData.variableData = (portspec, var_uuid, var_name)
            drag = QtGui.QDrag(self)
            drag.setMimeData(mimeData)
            drag.setHotSpot(pixmap.rect().bottomRight())
            drag.setPixmap(pixmap)
            drag.start(QtCore.Qt.MoveAction)

class QHoverVariableLabel(QtGui.QLabel):
    """
    QHoverVariableLabel is a QLabel that supports hover actions similar
    to a hot link
    """
    def __init__(self, var_name='', parent=None):
        """ QHoverVariableLabel(var_name:str,
                                parent: QWidget) -> QHoverVariableLabel
        Initialize the label with a variable name
        
        """
        QtGui.QLabel.__init__(self, parent)
        self.var_name = var_name
        self.setText(var_name)
        self.setAttribute(QtCore.Qt.WA_Hover)
        self.setCursor(QtCore.Qt.PointingHandCursor)
        self.setToolTip('Click to rename')
        self.palette().setColor(QtGui.QPalette.WindowText,
                                CurrentTheme.HOVER_DEFAULT_COLOR)

    def event(self, event):
        """ event(event: QEvent) -> Event Result
        Override to handle hover enter and leave events for hot links
        
        """
        if event.type()==QtCore.QEvent.HoverEnter:
            self.palette().setColor(QtGui.QPalette.WindowText,
                                    CurrentTheme.HOVER_SELECT_COLOR)
        if event.type()==QtCore.QEvent.HoverLeave:
            self.palette().setColor(QtGui.QPalette.WindowText,
                                    CurrentTheme.HOVER_DEFAULT_COLOR)
        return QtGui.QLabel.event(self, event)

    def mousePressEvent(self, event):
        """ mousePressEvent(event: QMouseEvent) -> None        
        If mouse click on the label, show up a dialog to change/add
        the variable name
        
        """
        if event.button()==QtCore.Qt.LeftButton:
            inputWidget = self.parent().parent()
            orig_var_name = inputWidget.var_name
            (text, ok) = QtGui.QInputDialog.getText(self,
                                                    'Set New Variable Name',
                                                    'Enter the new variable name',
                                                    QtGui.QLineEdit.Normal,
                                                    orig_var_name)
            var_name = str(text).strip()
            while ok and self.parent().parent().check_variable(var_name):
                msg =" This variable name is already being used.\
 Please enter a different variable name "
                (text, ok) = QtGui.QInputDialog.getText(self,
                                                        'Set New Variable Name',
                                                        msg,
                                                        QtGui.QLineEdit.Normal,
                                                        text)
                var_name = str(text).strip()
            if ok and var_name != orig_var_name:
                self.setText(var_name)
                inputWidget.renameVariable(var_name)




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from ast import literal_eval
import os
import sys
import urllib
import rpy2.robjects as robjects

from vistrails.core.modules.basic_modules import PathObject, new_constant
from vistrails.core.modules.vistrails_module import Module, ModuleError
from .widgets import RSourceConfigurationWidget, RFigureConfigurationWidget


# FIXME when rpy2 is installed on the path, we won't need this
old_sys_path = sys.path
sys.path.append(os.path.dirname(__file__))
sys.path = old_sys_path

r_temp_files = []

class TypeException(Exception):
    pass

def create_vector(v_list, desired_type=None):
    is_bool = True
    is_int = True
    is_float = True
    is_str = True
    for elt in v_list:
        if isinstance(elt, basestring):
            is_bool = False
            is_int = False
            is_float = False
        elif isinstance(elt, float):
            is_bool = False
            is_int = False
        elif isinstance(elt, (int, long)):
            is_bool = False
        else:
            is_bool = False
            is_int = False
            is_float = False
            is_str = False
            break
 
    if is_bool and (desired_type is None or desired_type == bool):
        return robjects.BoolVector(v_list)
    elif is_int and (desired_type is None or desired_type == int):
        res = [int(elt) for elt in v_list]
        return robjects.IntVector(res)
    elif is_float and (desired_type is None or desired_type == float):
        res = [float(elt) for elt in v_list]
        return robjects.FloatVector(res)
    elif is_str and (desired_type is None or desired_type == str):
        res = [str(elt) for elt in v_list]
        return robjects.StrVector(res)
    
    if desired_type is not None:
        raise TypeException("Cannot coerce vector to type '%s'" % desired_type)
    return robjects.RVector(v_list)

def vector_conv(v, desired_type=None):
    v_list = literal_eval(v)
    return create_vector(v_list, desired_type)

RVector = new_constant('RVector', staticmethod(vector_conv),
                       robjects.Vector([]),
                       staticmethod(lambda x: isinstance(x, robjects.Vector)))

def bool_vector_conv(v):
    return vector_conv(v, bool)

RBoolVector = new_constant('RBoolVector' , staticmethod(bool_vector_conv), 
                            robjects.BoolVector([]),
                            staticmethod(lambda x: isinstance(x, robjects.Vector)),
                            base_class=RVector)
                       
def int_vector_conv(v):
    return vector_conv(v, int)

RIntVector = new_constant('RIntVector' , staticmethod(int_vector_conv), 
                            robjects.IntVector([]),
                            staticmethod(lambda x: isinstance(x, robjects.Vector)),
                            base_class=RVector)

def float_vector_conv(v):
    return vector_conv(v, float)

RFloatVector = new_constant('RFloatVector' , staticmethod(float_vector_conv), 
                            robjects.FloatVector([]),
                            staticmethod(lambda x: isinstance(x, robjects.Vector)),
                            base_class=RVector)
                       
def str_vector_conv(v):
    return vector_conv(v, str)

RStrVector = new_constant('RStrVector' , staticmethod(str_vector_conv), 
                            robjects.StrVector([]),
                            staticmethod(lambda x: isinstance(x, robjects.Vector)),
                            base_class=RVector)

def array_conv(v):
    return robjects.r.array(vector_conv(v))

RArray = new_constant('RArray', staticmethod(array_conv),
                      robjects.r.array(),
                      staticmethod(lambda x: isinstance(x, robjects.RArray)),
                      base_class=RVector)

def create_matrix(v_list):
    vec_list = []
    nrow = 0
    ncol = -1
    for v_sublist in v_list:
        if ncol == -1:
            ncol = len(v_sublist)
        elif ncol != len(v_sublist):
            raise TypeException("Matrix must be rectangular")
        nrow += 1
        vec_list.extend(v_sublist)
    vec = create_vector(vec_list)
    return robjects.r.matrix(vec, nrow=nrow)
    
def matrix_conv(v):
    # should be a double list
    v_list = literal_eval(v)
    create_matrix(v_list)

def matrix_compute(self):
    if self.has_input('rvector'):
        rvector = self.get_input('rvector')
        nrows = self.get_input('nrows')
        self.set_output('value', robjects.r.matrix(rvector, nrow=nrows))
    else:
        RArray.compute(self)

RMatrix = new_constant('RMatrix', staticmethod(matrix_conv),
                       robjects.r.matrix(),
                       staticmethod(lambda x: isinstance(x, robjects.RArray)),
                       base_class=RArray,
                       compute=matrix_compute)
RMatrix._input_ports.extend([('rvector', '(Types|RVector)'), 
                             ('nrows', '(basic:Integer)')])

def create_list(v_dict):
    data_dict = {}
    for k,v in v_dict.iteritems():
        if isinstance(v, list):
            data_dict[k] = create_vector(v)
        elif isinstance(v, dict):
            data_dict[k] = create_list(v)
        else:
            data_dict[k] = v
    return robjects.r['list'](**data_dict)

def list_conv(v):
    v_dict = literal_eval(v)
    return create_list(v_dict)

RList = new_constant('RList', staticmethod(list_conv),
                     robjects.r.list(),
                     staticmethod(lambda x: isinstance(x, robjects.Vector)),
                     base_class=RVector)
# compute=list_compute)

def create_data_frame(v_dict):
    data_dict = {}
    for k,v in v_dict.iteritems():
        if isinstance(v, list):
            data_dict[k] = create_vector(v)
        elif isinstance(v, dict):
            data_dict[k] = create_data_frame(v)
        else:
            data_dict[k] = v
    return robjects.r['data.frame'](**data_dict)

def data_frame_conv(v):
    v_dict = literal_eval(v)
    return create_data_frame(v_dict)

RDataFrame = new_constant('RDataFrame', staticmethod(data_frame_conv),
                          robjects.r['data.frame'](),
                          staticmethod(lambda x: \
                                           isinstance(x, robjects.RDataFrame)))
                          
class RVectorFromList(Module):
    _input_ports = [('list', '(basic:List)')]
    _output_ports = [('rvector', '(Types|RVector)')]

    def compute(self):
        ilist = self.get_input('list')
        rvector = create_vector(ilist)
        self.set_output('rvector', rvector)

class ListFromRVector(Module):
    _input_ports = [('rvector', '(Types|RVector)')]
    _output_ports = [('list', '(basic:List)')]

    def compute(self):
        rvector = self.get_input('rvector')
        olist = list(rvector)
        self.set_output('list', olist)

class RMatrixFromNestedList(Module):
    _input_ports = [('list', '(basic:List)')]
    _output_ports = [('rmatrix', '(Types|RMatrix)')]

    def compute(self):
        ilist = self.get_input('list')
        rmatrix = create_matrix(ilist)
        self.set_output('rmatrix', rmatrix)

class NestedListFromRMatrix(Module):
    _input_ports = [('rmatrix', '(Types|RMatrix)')]
    _output_ports = [('list', '(basic:List)')]
    
    def compute(self):
        rmatrix = self.get_input('rmatrix')
        mlist = list(rmatrix)
        nrows = rmatrix.nrow
        ncols = len(mlist) // nrows
        olist = [] 
        for row in xrange(nrows):
            olist.append(mlist[row*ncols:(row+1)*ncols])
        self.set_output('list', olist)

class RDataFrameFromDict(Module):
    _input_ports = [('dict', '(basic:Dictionary)')]
    _output_ports = [('rdataframe', '(Types|RDataFrame)')]
    
    def compute(self):
        idict = self.get_input('dict')
        rdataframe = create_data_frame(idict)
        self.set_output('rdataframe', rdataframe)

class DictFromRDataFrame(Module):
    _input_ports = [('rdataframe','(Types|RDataFrame)')]
    _output_ports = [('dict', '(basic:Dictionary)')]

    def compute(self):
        rdataframe = self.get_input('rdataframe')
        colnames = list(rdataframe.colnames())
        odict = {}
        for i in xrange(len(rdataframe)):
            # FIXME !!! just assume that each row can be converted to a list!!!
            odict[colnames[i]] = list(rdataframe[i])
        self.set_output('dict', odict)

class RListFromDict(Module):
    # _input_ports = [('dict', '(basic:Dictionary)')]
    _input_ports = [('dict', '(basic:Module)')]
    _output_ports = [('rlist', '(Types|RList)')]
    
    def compute(self):
        idict = self.get_input('dict')
        rlist = create_list(idict)
        self.set_output('rlist', rlist)

class DictFromRList(Module):
    _input_ports = [('rlist', '(Types|RList)')]
    # _output_ports = [('dict', '(basic:Dictionary)')]
    _output_ports = [('dict', '(basic:Module)')]

    def compute(self):
        rlist = self.get_input('rlist')
        colnames = list(rlist.names)
        odict = {}
        for i in xrange(len(rlist)):
            # FIXME !!! just assume that each row can be converted to a list!!!
            # FIXME this may need to be a list of lists
            odict[colnames[i]] = list(rlist[i])
        self.set_output('dict', odict)

class RRead(Module):
    _input_ports = [('file', '(basic:File)'),
                    ('header', '(basic:Boolean)', True),
                    ('sep', '(basic:String)', True),
                    ('commentChar', '(basic:String)', True),
                    ('quote', '(basic:String)', True),
                    ('dec', '(basic:String)', True),
                    ('fill', '(basic:Boolean)', True)]
    _output_ports = [('rdataframe', '(Types|RDataFrame)')]

    def do_read(self, read_cmd):
        fname = self.get_input('file').name
        options_dict = {}
        for port in RRead._input_ports:
            if port[0] != 'file' and self.has_input(port):
                options_dict[port] = self.get_input(port)
        rdataframe = robjects.r[read_cmd](fname, **options_dict)
        self.set_output('rdataframe', rdataframe)

class RReadTable(RRead):
    def compute(self):
        self.do_read('read.table')

class RReadCSV(RRead):
    def compute(self):
        self.do_read('read.csv')

class RReadCSV2(RRead):
    def compute(self):
        self.do_read('read.csv2')
    
class RReadDelim(RRead):
    def compute(self):
        self.do_read('read.delim')

class RReadDelim2(RRead):
    def compute(self):
        self.do_read('read.delim2')

# class RSource(NotCacheable, Module):
class RSource(Module):
    _input_ports = [('source', '(basic:String)', True)]
    def run_code(self, code_str,
                 use_input=False,
                 use_output=False,
                 excluded_inputs=set(),
                 excluded_outputs=set()):
        """run_code runs a piece of code as a VisTrails module.
        use_input and use_output control whether to use the inputport
        and output port dictionary as local variables inside the
        execution."""
        import vistrails.core.packagemanager
        def fail(msg):
            raise ModuleError(self, msg)
        def cache_this():
            self.is_cacheable = lambda *args, **kwargs: True
        if use_input:
            inputDict = dict([(k, self.get_input(k))
                              for k in self.inputPorts
                              if k not in excluded_inputs])
            for k,v in inputDict.iteritems():
                robjects.globalEnv[k] = v
        robjects.r(code_str)
        if use_output:
            for k in self.outputPorts:
                if k not in excluded_outputs and k in robjects.globalEnv:
                    self.set_output(k, robjects.globalEnv[k])

    def run_file(self, fname, excluded_inputs=set(['source']), 
                 excluded_outputs=set()):
        f = open(fname, 'rU')
        code_str = f.readlines()
        f.close()
        self.run_code(code_str, use_input=True, use_output=True,
                      excluded_inputs=excluded_inputs,
                      excluded_outputs=excluded_outputs)

    def set_variable(self, name, value):
        robjects.globalEnv[name] = value

    def get_variable(self, name):
        # return robjects.globalEnv[name]
        return robjects.r(name)

    def chdir(self, dir):
        robjects.r('setwd("%s")' % dir)

    def compute(self):
        code_str = urllib.unquote(str(self.force_get_input('source', '')))
        self.run_code(code_str, use_input=True, use_output=True,
                      excluded_inputs=set(['source']))

class RFigure(RSource):
    _output_ports = [('imageFile', '(basic:File)')]
    def run_figure(self, code_str, graphics_dev, width, height, 
                   excluded_inputs=set(['source'])):
        fname = self.interpreter.filePool.create_file(prefix='vtr', suffix='.' + graphics_dev)
        r_temp_files.append(fname)
        robjects.r[graphics_dev](file=fname, width=width, height=height)
        self.run_code(code_str, use_input=True, 
                      excluded_inputs=excluded_inputs)
        robjects.r['dev.off']()
        image_file = PathObject(fname)
        self.set_output('imageFile', image_file)

    def run_figure_file(self, fname, graphics_dev, width, height, 
                        excluded_inputs=set(['source'])):
        f = open(fname)
        code_str = f.readlines()
        f.close()
        self.run_figure(code_str, graphics_dev, width, height, excluded_inputs)

class RSVGFigure(RFigure):
    def compute(self):
        code_str = \
            urllib.unquote(str(self.force_get_input('source', '')))
        RFigure.run_figure(self, code_str, 'svg', 4, 3)

class RPNGFigure(RFigure):
    def compute(self):
        code_str = \
            urllib.unquote(str(self.force_get_input('source', '')))
        RFigure.run_figure(self, code_str, 'png', 640, 480)

class RPDFFigure(RFigure):
    def compute(self):
        code_str = \
            urllib.unquote(str(self.force_get_input('source', '')))
        RFigure.run_figure(self, code_str, 'pdf', 4, 3)

class RFactor(Module):
    pass

_modules = {None: [(RSource, 
                    {'configureWidgetType': RSourceConfigurationWidget})],
            'Types': [RVector, 
                      RBoolVector,
                      RIntVector,
                      RFloatVector,
                      RStrVector,
                      RArray,
                      RMatrix,
                      RDataFrame,
                      RList,
                      RFactor],
            'Readers': [(RRead, {'abstract': True}),
                         RReadTable,
                         RReadCSV,
                         RReadCSV2,
                         RReadDelim,
                         RReadDelim2],
            'Converters': [RVectorFromList,
                           ListFromRVector,
                           RMatrixFromNestedList,
                           NestedListFromRMatrix,
                           RDataFrameFromDict,
                           DictFromRDataFrame,
                           RListFromDict,
                           DictFromRList],
            'Figures': [(RFigure, {'abstract': True}),
                        (RSVGFigure, {'configureWidgetType': \
                                          RFigureConfigurationWidget}),
                        (RPNGFigure, {'configureWidgetType': \
                                          RFigureConfigurationWidget}),
                        (RPDFFigure, {'configureWidgetType': \
                                          RFigureConfigurationWidget})]
            }

def finalize():
    for file in r_temp_files:
        os.remove(file)



from __future__ import absolute_import, division

import contextlib
import io
import logging
import os
import urllib

from vistrails.core import debug
from vistrails.core.bundles import py_import
from vistrails.core.modules.basic_modules import PathObject
from vistrails.core.modules.config import ModuleSettings
from vistrails.core.modules.vistrails_module import Module, ModuleError, \
    ModuleSuspended
from vistrails.core.vistrail.job import JobMixin


tej = py_import('tej', {'pip': 'tej'})


assert __name__.endswith('.init')
this_pkg = __name__[:-5]


class ServerLogger(tej.ServerLogger):
    def __init__(self):
        self.hidden = False
        tej.ServerLogger.__init__(self)

    @contextlib.contextmanager
    def hide_output(self):
        self.hidden = True
        try:
            yield
        finally:
            self.hidden = False

    def message(self, data):
        if self.hidden:
            debug.debug("tej server: %s" % data)
        else:
            debug.warning("tej server: %s" % data)

ServerLogger = ServerLogger()


class RemoteQueue(tej.RemoteQueue):
    def server_logger(self):
        return ServerLogger


class QueueCache(object):
    """A global cache of RemoteQueue objects.
    """
    def __init__(self):
        self._cache = {}

    def get(self, destination, queue):
        key = destination, queue
        if key in self._cache:
            return self._cache[key]
        else:
            queue = RemoteQueue(destination, queue)
            self._cache[key] = queue
            return queue

QueueCache = QueueCache()


class Queue(Module):
    """A connection to a queue on a remote server.

    `hostname` can be a hostname or a full destination in the format:
    ``[ssh://][user@]server[:port]``, e.g. ``vistrails@nyu.edu``.
    """
    _input_ports = [('hostname', '(basic:String)'),
                    ('username', '(basic:String)',
                     {'optional': True}),
                    ('port', '(basic:Integer)',
                     {'optional': True, 'defaults': "['22']"}),
                    ('queue', '(basic:String)',
                     {'optional': True, 'defaults': "['~/.tej']"})]
    _output_ports = [('queue', '(org.vistrails.extra.tej:Queue)')]

    def compute(self):
        destination_str = self.get_input('hostname')
        if self.has_input('username') or self.has_input('port'):
            destination = {'hostname': destination_str,
                           'username': self.get_input('username'),
                           'port': self.get_input('port')}
            destination_str = tej.destination_as_string(destination)

        queue = self.get_input('queue')
        self.set_output('queue', QueueCache.get(destination_str, queue))


class RemoteJob(object):
    def __init__(self, queue, job_id):
        self.queue = queue
        self.job_id = job_id

    def finished(self):
        try:
            with ServerLogger.hide_output():
                status, target, arg = self.queue.status(self.job_id)
        except tej.JobNotFound:
            # We signal that we are done
            # The Module will raise an error on resume, as intended
            return True
        return status == tej.RemoteQueue.JOB_DONE


class Job(Module):
    """A reference to a job in a queue.

    Objects represented by this type only represent completed jobs, since else,
    the creating module would have failed/suspended.

    You probably won't use this module directly since it references a
    pre-existing job by name.
    """
    _input_ports = [('id', '(basic:String)'),
                    ('queue', Queue)]
    _output_ports = [('job', '(org.vistrails.extra.tej:Job)'),
                     ('exitcode', '(basic:Integer)')]

    def compute(self):
        queue = self.get_input('queue')
        job_id = self.get_input('id')

        # Check job status
        try:
            with ServerLogger.hide_output():
                status, target, arg = queue.status(job_id)
        except tej.JobNotFound:
            raise ModuleError(self, "Job not found")

        # Create job object
        job = RemoteJob(queue=queue, job_id=job_id)

        if status == tej.RemoteQueue.JOB_DONE:
            self.set_output('job', job)
            self.set_output('exitcode', int(arg))
        elif status == tej.RemoteQueue.JOB_RUNNING:
            raise ModuleSuspended(self, "Remote job is running",
                                  handle=job)
        else:
            raise ModuleError(self, "Invalid job status %r" % status)


class BaseSubmitJob(JobMixin, Module):
    """Starts a job on a server.

    Thanks to the suspension/job tracking mechanism, this module does much more
    than start a job. If the job is running, it will suspend again. If the job
    is finished, you can obtain files from it.
    """
    _settings = ModuleSettings(abstract=True)
    _input_ports = [('queue', Queue)]
    _output_ports = [('job', '(org.vistrails.extra.tej:Job)'),
                     ('exitcode', '(basic:Integer)')]

    def make_id(self):
        """Makes a default identifier, using the pipeline signature.

        Unused if we have an explicit identifier.
        """
        if not hasattr(self, 'signature'):
            raise ModuleError(self,
                              "No explicit job ID and module has no signature")
        return "vistrails_module_%s" % self.signature

    def job_read_inputs(self):
        """Reads the input ports.
        """
        return {'destination': self.get_input('queue').destination_string,
                'queue': str(self.get_input('queue').queue),
                'job_id': self.make_id()}

    def job_start(self, params):
        """Submits a job.

        Reimplement in subclasses to actually submit a job.
        """
        raise NotImplementedError

    def job_get_handle(self, params):
        """Gets a RemoteJob object to monitor a runnning job.
        """
        queue = QueueCache.get(params['destination'], params['queue'])
        return RemoteJob(queue, params['job_id'])

    def job_finish(self, params):
        """Finishes job.

        Gets the exit code from the server.
        """
        queue = QueueCache.get(params['destination'], params['queue'])
        status, target, arg = queue.status(params['job_id'])
        assert status == tej.RemoteQueue.JOB_DONE
        params['exitcode'] = int(arg)
        return params

    def job_set_results(self, params):
        """Sets the output ports once the job is finished.
        """
        queue = QueueCache.get(params['destination'], params['queue'])
        self.set_output('exitcode', params['exitcode'])
        self.set_output('job', RemoteJob(queue, params['job_id']))


class SubmitJob(BaseSubmitJob):
    """Submits a generic job (a directory).
    """
    _input_ports = [('job', '(basic:Directory)'),
                    ('script', '(basic:String)',
                     {'optional': True, 'defaults': "['start.sh']"})]

    def job_start(self, params):
        """Sends the directory and submits the job.
        """
        queue = QueueCache.get(params['destination'], params['queue'])

        # First, check if job already exists
        try:
            with ServerLogger.hide_output():
                queue.status(params['job_id'])
        except (tej.JobNotFound, tej.QueueDoesntExist):
            pass
        else:
            return params

        # Alright, submit a new job
        queue.submit(params['job_id'],
                     self.get_input('job').name,
                     self.get_input('script'))
        return params


class SubmitShellJob(BaseSubmitJob):
    """Submits a shell script.
    """
    _settings = ModuleSettings(configure_widget=(
            '%s.widgets' % this_pkg, 'ShellSourceConfigurationWidget'))
    _input_ports = [('source', '(basic:String)')]
    _output_ports = [('stderr', '(basic:File)'),
                     ('stdout', '(basic:File)')]

    _job_interpreter = '/bin/sh'

    def job_start(self, params):
        """Creates a temporary job with the given source, upload and submit it.
        """
        queue = QueueCache.get(params['destination'], params['queue'])

        # First, check if job already exists
        try:
            with ServerLogger.hide_output():
                queue.status(params['job_id'])
        except (tej.JobNotFound, tej.QueueDoesntExist):
            pass
        else:
            return params

        # Alright, submit a new job
        directory = self.interpreter.filePool.create_directory(
                prefix='vt_tmp_shelljob_').name
        # We use io.open() here because we could be writing scripts on Windows
        # before uploading them to a POSIX server
        source = urllib.unquote(self.get_input('source'))
        if isinstance(source, bytes):
            kwargs = {'mode': 'wb'}
        else:
            kwargs = {'mode': 'w', 'newline': '\n'}
        with io.open(os.path.join(directory, 'vistrails_source.sh'),
                     **kwargs) as fp:
            fp.write(source)
        with io.open(os.path.join(directory, 'start.sh'), 'w',
                     newline='\n') as fp:
            fp.write(u'%s '
                     u'vistrails_source.sh '
                     u'>_stdout.txt '
                     u'2>_stderr.txt\n' % self._job_interpreter)

        queue.submit(params['job_id'], directory)

        return params

    def job_set_results(self, params):
        """Gets stderr and stdout.
        """
        super(SubmitShellJob, self).job_set_results(params)

        temp_dir = self.interpreter.filePool.create_directory(
                prefix='vt_tmp_shelljobout_').name
        queue = QueueCache.get(params['destination'], params['queue'])
        queue.download(params['job_id'], ['_stderr.txt', '_stdout.txt'],
                       directory=temp_dir)
        self.set_output('stderr',
                        PathObject(os.path.join(temp_dir, '_stderr.txt')))
        self.set_output('stdout',
                        PathObject(os.path.join(temp_dir, '_stdout.txt')))


class DownloadFile(Module):
    """Downloads a file from a remote job.
    """
    _input_ports = [('job', Job),
                    ('filename', '(basic:String)')]
    _output_ports = [('file', '(basic:File)')]

    def compute(self):
        job = self.get_input('job')
        assert isinstance(job, RemoteJob)

        destination = self.interpreter.filePool.create_file(
                prefix='vt_tmp_shelljobout_')
        job.queue.download(job.job_id,
                           self.get_input('filename'),
                           destination=destination.name,
                           recursive=False)

        self.set_output('file', destination)


class DownloadDirectory(Module):
    """Downloads a directory from a remote job.
    """
    _input_ports = [('job', Job),
                    ('pathname', '(basic:String)')]
    _output_ports = [('directory', '(basic:Directory)')]

    def compute(self):
        job = self.get_input('job')
        assert isinstance(job, RemoteJob)

        destination = self.interpreter.filePool.create_directory(
                prefix='vt_tmp_shelljobout_').name
        target = os.path.join(destination, 'dir')
        job.queue.download(job.job_id,
                           self.get_input('pathname'),
                           destination=target,
                           recursive=True)

        self.set_output('directory', PathObject(target))


_tej_log_handler = None


class _VisTrailsTejLogHandler(logging.Handler):
    def emit(self, record):
        msg = "tej: %s" % self.format(record)
        if record.levelno >= logging.CRITICAL:
            debug.critical(msg)
        elif record.levelno >= logging.WARNING:
            debug.warning(msg)
        elif record.levelno >= logging.INFO:
            debug.log(msg)
        else:
            debug.debug(msg)


def initialize():
    # Route tej log messages to VisTrails
    global _tej_log_handler
    _tej_log_handler = _VisTrailsTejLogHandler()
    tej_logger = logging.getLogger('tej')
    tej_logger.propagate = False
    tej_logger.addHandler(_tej_log_handler)
    tej_logger.setLevel(logging.DEBUG)


def finalize():
    # Unregister logging handler; useful for reloading support
    global _tej_log_handler
    tej_logger = logging.getLogger('tej')
    if _tej_log_handler is not None:
        tej_logger.removeHandler(_tej_log_handler)
        _tej_log_handler = None
    tej_logger.propagate = True


_modules = [Queue,
            Job, BaseSubmitJob, SubmitJob, SubmitShellJob,
            DownloadFile, DownloadDirectory]
